#!/usr/bin/python -tt
# -*- coding: utf-8 -*-
# =============================================================================
# File      : paths.py
# Creation  : 29 Mar 2018
# Time-stamp: <Don 2018-07-26 08:57 juergen>
#
# Copyright (c) 2018 Jürgen Hackl <hackl@ibi.baug.ethz.ch>
#               http://www.ibi.ethz.ch
# $Id$
#
# Description : Base class for paths
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# =============================================================================

import pickle
from copy import deepcopy
from cnet import config, logger
from cnet.classes.network import Network, Node, Edge
from cnet.utils.exceptions import CnetError, CnetNotImplemented
log = logger(__name__)

MAX_NAME_LENGTH = 5


class Path(Network):
    """Base class for paths.

    A Path stores consecutive nodes and edges with optional data, or attributes.

    Instances of this class capture a path that can be directed, un-directed,
    un-weighted or weighted. The path class is a subclass of a network.

    Nodes are defined as :py:class:`Node` and edges are defined as
    :py:class:`Edge`

    Parameters
    ----------
    nodes : list of node ids or Nodes
        The parameter nodes must be a list with node ids or node
        objects. Every node within the list should have a unique id.
        The id is converted to a string value and is used as a key value for
        all dict which saving node objects.

    directed : Boole, optional  (default = True)
        Specifies if a network contains directed edges, i.e u->v or
        undirected edges i.d. u<->v. Per default the path is assumed to
        be directed.

    name : string, optional (default = '')
        An optional name for the path. If no name is assigned the network is
        called after the assigned nodes. e.g. if the path has nodes 'a', 'b' and
        'c', the network is named 'a-b-c'.

    separator : string, optional (default = '-')
        The separator used to separate the nodes in the node name.

    attr : keyword arguments, optional (default= no attributes)
        Attributes to add to the path as key=value pairs.

    Attributes
    ----------
    name : string
        Name of the path. If no name is assigned the network is called after the
        assigned nodes. e.g. if the path has nodes 'a', 'b' and 'c', the network
        is named 'a-b-c'. The maximum length of the name is defined by the
        constant `MAX_NAME_LENGTH`, which is per default 5. E.g. if the path has
        7 nodes (a,b,c,d,e,f,g) the name of the path is 'a-b-c-d-...-g'. Please,
        note the name of a path is NOT an unique identifier!

    id : string
        The identifier of the :py:class:`Path` object. Every path should
        have a unique id. The id is converted to a string value. If no attribute
        id is defined, the id is created by the nodes in the path. e.g. a path
        with such as (a,b,c,d,e,f,a) has the id 'a-b-c-d-e-f-g-a'.

    Examples
    --------
    Create an empty path with no nodes and no edges.

    >>> p = cn.Path()

    Some properties of the path are: the name, if directed or the shape

    >>> p.name = 'my test path'
    >>> p.name
    my test path

    Adding a singe new node to the network.

    >>> p.add_node('a',color='red')

    Adding an existing node object to the network.

    >>> b = cn.Node('b',color='green')
    >>> p.add_node(b)

    An edge was generated between a and b

    >>> p.name
    a-b

    Adding a singe new edge to the path.

    >>> p.add_edge('cd','c','d', length = 10)

    Adding an existing edge object to the path.

    >>> e = cn.Edge('da','d','a', length = 5)
    >>> p.add_edge(e)
    >>> p.name
    a-b-c-d-a

    See Also
    --------
    SpatialPath

    """

    def __init__(self, nodes=None, directed=True, separator='-', **attr):
        """Initialize a path with direction, name and attributes.

        Parameters
        ----------
        nodes : list of node ids or Nodes, optional (default = None)
            The parameter nodes must be a list with node ids or node
            objects. Every node within the list should have a unique id.
            The id is converted to a string value and is used as a key value for
            all dict which saving node objects.

        directed : Boole, optional  (default = True)
            Specifies if a network contains directed edges, i.e u->v or
            undirected edges i.d. u<->v. Per default the path is assumed to
            be directed.

        name : string, optional (default = '')
            An optional name for the path. If no name is assigned the network is
            called after the assigned nodes. e.g. if the path has nodes 'a', 'b'
            and 'c', the network is named 'a-b-c'.

        separator : string, optional (default = '-')
            The separator used to separate the nodes in the node name.

        attr : keyword arguments, optional (default= no attributes)
            Attributes to add to the path as key=value pairs.

        """
        # Classes of the Path objects
        # TODO: Probably there is a better solution to have different Path
        # classes for different Path sub classes
        self._path_class()

        # an modified list object containing node and edge objects
        self.path = PathList()

        # initializing the separator
        self.separator = separator
        # initializing the parent network class
        super().__init__(directed=directed, **attr)

        # assign nodes to the path
        if isinstance(nodes, list):
            super().add_nodes_from(nodes)

    def _path_class(self):
        """Internal function to assign different Path classes."""
        self.PathClass = Path

    def __repr__(self):
        """Return the description of the path (see :meth:`_desc`) with the name
        of the path."""
        return '<{} object {} at 0x{}x>'.format(self._desc(), self.name, id(self))

    def _desc(self):
        """Return a string *Path()*."""
        return '{}'.format(self.__class__.__name__)

    def __len__(self):
        """Returns the number of nodes in the path"""
        return len(self.path)

    def __eq__(self, other):
        """Returns True if two paths are equal.

        Here equality is defined by the number and order of nodes in the path!

        """
        return self.id == other.id

    def __hash__(self):
        """Returns the unique hash of the path.

        Here the hash value is defined by the string of nodes in the path!

        """
        return hash(self.id)

    @property
    def name(self):
        """Returns the name of the path.

        Name of the path. If no name is assigned the network is called after the
        assigned nodes. e.g. if the path has nodes 'a', 'b' and 'c', the network
        is named 'a-b-c'. The maximum length of the name is defined by the
        constant `MAX_NAME_LENGTH`, which is per default 5. E.g. if the path has
        7 nodes (a,b,c,d,e,f,g) the name of the path is 'a-b-c-d-...-g'. Please,
        note the name of a path is NOT an unique identifier!

        Returns
        -------
        name : string
            Returns the name of the path as a string.

        """
        max_name_length = MAX_NAME_LENGTH
        if len(self) > max_name_length:
            _name = self.separator.join(self.path[0:-2][0:max_name_length]) \
                + self.separator + '...' + self.separator + self.path[-1]
        else:
            _name = self.separator.join(self.path)

        return self.attributes.get('name', _name)

    @property
    def full_name(self):
        '''Returns the full name of the path

        Name of the path. If no name is assigned the network is called after the
        assigned nodes. e.g. if the path has nodes 'a', 'b' and 'c', the network
        is named 'a-b-c'.

        Returns
        -------
        name : string
            Returns the name of the path as a string.

        '''
        _name = self.separator.join(self.path)

        return self.attributes.get('name', _name)

    @property
    def id(self):
        """Returns the id of the path.

        Id of the path. If no id is assigned the path is called after the
        assigned nodes. e.g. if the path has nodes 'a', 'b' and 'c', the id is
        'a-b-c'.

        Returns
        -------
        id : string
            Returns the id of the path as a string.

        """
        _id = self.separator.join(self.path)

        return self.attributes.get('id', _id)

    def weight(self, weight='weight'):
        """Returns the weight of the path.

        Per default the attribute with the key 'weight' is used as
        weight. Should there be no such attribute, a new one will be crated with
        weight = 1.0.

        If an other attribute should be used as weight, the option weight has to
        be changed.

        If a weight is assigned but for calculation a weight of 1.0 is needed,
        the weight can be disabled with False or None.

        Parameters
        ----------
        weight : str, optional (default = 'weight')
            The weight parameter defines which attribute is used as weight. Per
            default the attribute 'weight' is used. If None or False is chosen,
            the weight will be 1.0. Also any other attribute of the path can be
            used as a weight

        Returns
        -------
        weight : attribute
            Returns the attribute value associated with the keyword.

        Example
        -------
        Create new empty path and get the weight.

        >>> p = nc.Path()
        >>> p.weight()
        1.0

        Change the weight.

        >>> p['weight'] = 4
        >>> p.weight()
        4
        >>> p.weight(False)
        1.0

        Add an attribute and use this as weight.

        >>> p['length'] = 5
        >>> p.weight('length')
        5

        """
        if weight is None:
            weight = False
        if not weight:
            return 1.0
        elif isinstance(weight, str) and weight != 'weight':
            return self.attributes.get(weight, 1.0)
        else:
            return self.attributes.get('weight', 1.0)

    def summary(self):
        """Returns a summary of the path.

        The summary contains the name, the used path class, if it is directed
        or not, the number of unique nodes and unique edges, and the number of
        nodes in the path.

        Since a path can multiple times pass the same node and edge objects, the
        length of the path (i.e. the consecutive nodes) might be larger then the
        number of unique nodes.

        If logging is enabled (see config), the summary is written to the log
        file and showed as information on in the terminal. If logging is not
        enabled, the function will return a string with the information, which
        can be printed to the console.

        """
        summary = [
            'Name:\t\t\t{}\n'.format(self.name),
            'Type:\t\t\t{}\n'.format(self.__class__.__name__),
            'Directed:\t\t{}\n'.format(str(self.directed)),
            'Number of unique nodes:\t{}\n'.format(self.number_of_nodes()),
            'Number of unique edges:\t{}'.format(self.number_of_edges()),
            'Path length (# nodes):\t{}'.format(len(self))
        ]
        if config.logging.enabled:
            for line in summary:
                log.info(line.rstrip())
        else:
            return ''.join(summary)

    def add_node(self, n, **attr):
        """Add a single node n and update node attributes.

        Add a single node to the path object. If a second node is added,
        automatically an edge is created between these two nodes.

        Parameters
        ----------
        u : node id or Node
            The parameter u is the identifier (id) for the node. Every node
            should have a unique id. The id is converted to a string value and
            is used as a key value for all dict which saving node objects.

        attr : keyword arguments, optional (default= no attributes)
            Attributes to add to node as key=value pairs.

        Example
        -------
        Adding a singe new node to the network.

        >>> p = cn.Path()
        >>> p.add_node('a',color='red')

        Adding an existing node object to the network.

        >>> b = cn.Node('b',color='green')
        >>> p.add_node(b)

        An edge was generated between a and b

        >>> p.name
        a-b

        Note
        ----
        If the same node is entered twice in a row an warning will occur and the
        second node entry will be neglected.

        """
        # check if n is not a Node object
        if not isinstance(n, self.NodeClass):
            _node = self.NodeClass(n, **attr)
        else:
            _node = n
            _node.update(**attr)

        # check if node is the first node in the path
        if len(self) < 1:
            self.nodes[_node.id] = _node
            self.path.append(_node.id)
        # check if node is similar to previous node in the path
        elif self.path[-1] == _node.id:
            log.warn('The predecessor node is similar to node "{}"!'
                     ' No node was added to the path!'.format(_node.id))
            return None
        # check if node is connected to previous node in the path
        # elif self.has_edge((self.path[-1],_node.id)):
        #     log.debug('Can this happen? If you see this msg then yes :(')
        else:
            u = self.path[-1]
            self.nodes[_node.id] = _node
            self.path.append(_node.id)
            super().add_edge(str(u)+self.separator+str(_node.id), u, _node.id)
            self.path.edges.append(str(u)+self.separator+str(_node.id))

    def add_edge(self, e, u=None, v=None, **attr):
        """Add an edge e between u and v to the path.

        Add an single edge to the path object. The edge must be connected to the
        predecessor edge.

        Parameters
        ----------
        e : edge id or Edge object
            The parameter e is the identifier (id) for the edge. Every edge
            should have a unique id. The id is converted to a string value and
            is used as a key value for all dict which saving edge objects.
            If e is already an edge object, no nodes have to be assigned. If e
            is a edge id, u and v have to be defined. If not an error will occur.

        u : node id or Node, optional (default = None)
            This parameter defines the origin of the edge (if directed),
            i.e. u->v. A node id can be entered, in this case a new Node will be
            created (see py:class:Node), also an existing node can be used, here
            the attributes and properties of the node objects are used.

        v : node id or Node, optional (default = None)
            This parameter defines the destination of the edge (if directed)
            i.e. u->v. A node id can be entered, in this case a new Node will be
            created (see py:class:Node), also an existing node can be used, here
            the attributes and properties of the node objects are used.

        attr : keyword arguments, optional (default= no attributes)
            Attributes to add to edge as key=value pairs.

        Example
        -------
        Adding a singe new edge to the path.

        >>> p = cn.Path()
        >>> p.add_edge('ab','a','b', length = 10)

        Adding an existing edge object to the path.

        >>> e = cn.Edge('bc','b','c', length = 5)
        >>> p.add_edge(e)

        Note
        ----

        If an edge id is added without nodes, an error will be raised and stop
        the code.

        If the edge is already defined in the network. The new edge will not
        be added to the network, instead a warning will be printed.

        """
        # check if e is not an Edge object
        if not isinstance(e, self.EdgeClass):
            if u is not None and v is not None:
                if u in self.nodes:
                    u = self.nodes[u]
                if v in self.nodes:
                    v = self.nodes[v]
                _edge = self.EdgeClass(e, u, v, **attr)
            else:
                log.error('No nodes are defined for new edge "{}!"'.format(e))
                raise CnetError
        else:
            _edge = e
            _edge.update(**attr)

        # check if nodes are already in the path
        if _edge.u.id not in self.nodes:
            self.nodes[_edge.u.id] = _edge.u
        if _edge.v.id not in self.nodes:
            self.nodes[_edge.v.id] = _edge.v

        # check if node is the first node in the path
        if len(self) < 1:
            self.path.append(_edge.u.id)
            self.path.append(_edge.v.id)
            self._add_edge(_edge)
        # check if edge is connected to previous edge in the path
        elif self.path[-1] == _edge.u.id:
            self.path.append(_edge.v.id)
            self._add_edge(_edge)
        elif not self.directed and \
                (self.path[-1] == _edge.u.id or self.path[-1] == _edge.v.id):
            if self.path[-1] == _edge.u.id:
                self.path.append(_edge.v.id)
                self._add_edge(_edge)
            else:
                self.path.append(_edge.u.id)
                self._add_edge(_edge)

        else:
            log.error('Edge "{}" with nodes "({},{})", is not connected to the'
                      ' previous edge with nodes "({},{})!"'
                      ''.format(_edge.id, _edge.u.id, _edge.v.id, self.path[-2], self.path[-1]))

            raise CnetError

    def _add_edge(self, _edge):
        """Help function to add an edge."""
        self.edges[_edge.id] = _edge
        self.path.edges.append(_edge.id)
        self.nodes[_edge.u.id].heads.add((_edge.id, 0))
        self.nodes[_edge.v.id].tails.add((_edge.id, 0))

        if not self.directed:
            self.nodes[_edge.u.id].tails.add((_edge.id, 1))
            self.nodes[_edge.v.id].heads.add((_edge.id, 1))

    def remove_node(self, n):
        """Remove node n and all adjacent edges."""
        raise CnetNotImplemented

    def remove_nodes_from(self, nodes):
        """Remove multiple nodes."""
        raise CnetNotImplemented

    def remove_edge(self, e):
        """Remove the edge e between u and v."""
        raise CnetNotImplemented

    def remove_edges_from(self, edges):
        """Remove all edges specified in edges."""
        raise CnetNotImplemented

    def has_subpath(self, subpath, mode='nodes'):
        """Return True if the path has a sub path.

        Parameters
        ----------
        subpath : list of node or edge ids or a path object
            The sub path is a list of consecutive nodes or edges describing the
            path.

        mode : string, 'nodes' or 'edges', optional (default = 'nodes')
            The mode defines how the sup path is defined, i.e. as a list of node
            ids or a list of edge ids. If the sub path is a path object, the
            mode does not matter.

        Examples:
        ---------
        >>> p = cn.Path(['a','b','c','d'])
        >>> p.has_subpath(['a','b'])
        True

        Test other path object

        >>> q = cn.Path(['d','e'])
        >>> p.has_subpath(q)
        False

        Test edges.

        >>> p.has_subpath(['a-b','b-c'], mode='edges')
        True

        """
        if isinstance(subpath, self.PathClass):
            _subpath = subpath.path
        else:
            # TODO: check also if nodes and edges are Node or Edge classes.
            if mode == 'edges':
                try:
                    _subpath = self.edges_to_path(subpath)
                except:
                    return False
            else:  # default mode == 'nodes'
                _subpath = subpath
        # check if all elements of the sub path are in the path
        if set(_subpath).issubset(set(self.path)):
            # check the order of the elements
            # consider also elements which appear multiple times in the path
            idx = [i for i, x in enumerate(self.path) if x == _subpath[0]]
            return any(all(self._check_path(i+j, v) for j, v in
                           enumerate(_subpath)) for i in idx)
        else:
            return False

    def _check_path(self, index, value):
        """Help function to check of the index exist"""
        try:
            return self.path[index] == value
        except:
            return False

    def edges_to_path(self, edges):
        """Returns a list of node ids representing the path.

        Parameters
        ----------
        edges : list of edge ids
            The path defined as a list of consecutive edges.

        Returns
        -------
        path : list of node ids
            Returns a list of node ids representing the path object

        Note
        ----
        An error will be raised if there is no corresponding path.

        """
        if isinstance(edges, str):
            edges = [edges]

        # TODO: Something is wrong here
        e2n = self.edge_to_nodes_map()
        try:
            # add first edge
            _path = [e2n[edges[0]][0], e2n[edges[0]][1]]
            # add remaining edges
            _path += [e2n[edges[e]][1] for e in range(1, len(edges))]
            # return path
            return _path
        except Exception as error:
            log.error('The edges "{}" could not be mapped to the path "{}"!'
                      ''.format('-'.join(edges), self.name))
            raise

    def subpath(self, subpath, mode='nodes'):
        """Returns a sup path of the path.

        Parameters
        ----------
        subpath : list of node or edge ids
            The sub path is a list of consecutive nodes or edges describing the
            path.

        mode : string, 'nodes' or 'edges', optional (default = 'nodes')
            The mode defines how the sup path is defined, i.e. as a list of node
            ids or a list of edge ids. If the sub path is a path object, the
            mode does not matter.

        Returns
        -------
        subpath : Path
            Returns a Path object containing the sub path and the attributes of
            the parent path object.

        Examples
        --------
        >>> p = cn.Path(['a','b','c','d'])
        >>> q = p.subpath(['a','b','c'])
        >>> q.name
        a-b-c

        Sub path from edge list

        >>> q = p.subpath(['c-d'])
        >>> q.name
        c-d

        """
        # check if the sub path is in the path
        if self.has_subpath(subpath, mode=mode):
            if mode == 'edges':
                _subpath = self.edges_to_path(subpath)
            else:  # default mode == 'nodes'
                _subpath = subpath
        else:
            log.error('Path "{}" has not sub path "{}"!'
                      ''.format(self.name, '-'.join(subpath)))
            raise ValueError

        # create a new path opject
        subpath = self.PathClass()
        # copy the attributes of the parent path
        subpath.update(**self.attributes)

        # get node dict to map the edges
        n2e = super().nodes_to_edges_map()

        # go thought the sub path and add the edges
        for i in range(len(_subpath)-1):
            _e = self.edges[n2e[_subpath[i], _subpath[i+1]][0]]
            subpath.add_edge(_e)

        return subpath

    def subpaths(self, min_length=None, max_length=None, include_path=False):
        """Returns a paths object with all sub paths.

        Parameters
        ----------

        min_length : int, optional (default = None)
            Parameter which defines the minimum length of the sub paths. This
            parameter has to be smaller then the maximum length parameter.

        max_length : int, optional (default = None)
            Parameter which defines the maximum  length of the sub paths. This
            parameter has to be greater then the minimum length parameter. If
            the parameter is also greater then the maximum length of the path,
            the maximum path length is used instead.

        include_path : Boole, optional (default = Flase)
            If this option is enabled also the current path is added as a
            sub path of it self.

        Returns
        -------
        subpaths : Paths
            Returns a paths object containing all the sub paths fulfilling the
            length criteria.

        Examples
        --------
        >>> p = cn.Path(['a','b','c','d','e'])
        >>> P = p.subpaths()
        >>> for q in P:
        >>>     print(q.name)
        a-b
        b-c
        c-d
        d-e
        a-b-c
        b-c-d
        c-d-e
        a-b-c-d
        b-c-d-e

        >>> P = p.subpaths(min_length = 3, max_length = 3)
        >>> for q in P:
        >>>     print(q.name)
        a-b-c
        b-c-d
        c-d-e

        Note
        ----
        If the minimum length is larger then the maximum length, an error will
        raised. If the maximum length is larger then the actual path length, the
        actual path length will be used.

        """
        if min_length is None:
            min_length = 1
        if max_length is None:
            max_length = len(self)-1
        if min_length > max_length:
            log.error('The minimum path length {} must be smaller then the '
                      'maximum path length {}!'.format(min_length, max_length))
            raise ValueError
        max_length = min(max_length, len(self)-1)
        min_length = max(min_length, 2)

        subpaths = Paths(name='sub paths of '+self.name)
        for i in range(min_length-1, max_length):
            for j in range(len(self)-i):
                subpaths.add_path(self.subpath(self.path[j:j+i+1]))

        if len(subpaths) == 0 and len(self) == 2:
            subpaths.add_path(self.copy())
        elif len(self) <= max_length+1 and include_path:
            subpaths.add_path(self.copy())

        return subpaths

    def shared_paths(self, other):
        """Returns the shared path with an other path object"""
        pass


class PathList(list):
    """An extended class of list to save the nodes and edges of the path."""

    def __init__(self):
        self.edges = []
        self.nodes = []

    def append(self, val):
        """Append value to the path list and the node list."""
        self.insert(len(self), val)
        self.nodes.append(val)


class Paths(object):
    """Basic class to store and handle multiple Path classes.

    Parameters
    ----------
    paths : list of :py:class:`Path` objects, optional (default = None)
        The parameter paths must be a list with :py:class:`Path` objects.

    attr : keyword arguments, optional (default= no attributes)
        Attributes to add to the paths as key=value pairs.

    Attributes
    ----------
    name : string, optional (default='')
        An optional name for the network.

    """

    def __init__(self, paths=None, **attr):
        """Initialize the paths object with name and attributes.

        Parameters
        ----------
        paths : list of :py:class:`Path` objects, optional (default = None)
            The parameter paths must be a list with :py:class:`Path` objects.

        attr : keyword arguments, optional (default= no attributes)
            Attributes to add to the paths as key=value pairs.

        Attributes
        ----------
        name : string, optional (default='')
            An optional name for the network.
        """
        # dictionary for the path objects
        self.paths = []
        # dictionary for paths attributes
        self.attributes = {}
        # add attributes to the paths
        self.attributes.update(attr)

        # assign path to the path list
        if isinstance(paths, list):
            self.add_paths_from(paths)

    @property
    def name(self):
        """Return the name of the paths if defined, else an empty space."""
        return self.attributes.get('name', '')

    @name.setter
    def name(self, s):
        """Set the name of the paths."""
        self.attributes['name'] = s

    def __repr__(self):
        """Return the description of the paths (see :meth:`_desc`) with the id
        of the paths."""
        return '<{} object {} at 0x{}x>'.format(self._desc(), self.name, id(self))

    def _desc(self):
        """Return a string *Paths()*."""
        return '{}'.format(self.__class__.__name__)

    def __len__(self):
        """Returns the number of paths"""
        return len(self.paths)

    def __getitem__(self, index):
        """Returns the path with the assigned index, or the attribute with the
        assigned keyword."""
        if isinstance(index, int):
            return self.paths[index]
        elif isinstance(index, str):
            return self.attributes[index]
        else:
            log.error('Index is not defined properly. Please use integer to'
                      'get the path at the index, or a string value to get'
                      'the attribute of the Paths object')
            raise CnetError

    def __iter__(self):
        """Iterating trough the path objects."""
        return (path for path in self.paths)

    def update(self, **attr):
        """Update the attributes of the paths.

        Parameters
        ----------
        attr : keyword arguments, optional (default= no attributes)
            Attributes to add or update for the paths as key=value pairs.

        Examples
        --------
        Update attributes.

        >>> P = nc.Paths(type = 'road travelers')
        >>> P.update(type = 'rail travelers', location = 'Swizerland')

        """
        self.attributes.update(attr)

    def add_path(self, p):
        """Add a single path to the path list.

        Parameters
        ----------
        p : Path
            Add a path object to the list of paths

        Examples
        --------
        >>> P = cn.Paths()
        >>> P.add_path(cn.Path(['a','b','c']))

        """
        if isinstance(p, Path):
            # if p is not None:
            self.paths.append(p)

    def add_paths_from(self, paths):
        """Add multiple paths from a list.

        Parameters
        ----------
        paths : list of path objects
            Add paths from a list of path objects.

        Examples
        --------
        >>> P = cn.Paths()
        >>> paths = [cn.Path(['a','b','c']),cn.Path(['b','c'])]
        >>> P.add_path(paths)

        """
        for p in paths:
            if isinstance(p, Path):
                self.paths.append(p)

    def summary(self):
        """Returns a summary of the paths.

        The summary contains the name, the used paths class, and the number of paths.

        If logging is enabled (see config), the summary is written to the log
        file and showed as information on in the terminal. If logging is not
        enabled, the function will return a string with the information, which
        can be printed to the console.

        """
        summary = [
            'Name:\t\t\t{}\n'.format(self.name),
            'Type:\t\t\t{}\n'.format(self.__class__.__name__),
            'Number of paths:\t{}'.format(str(len(self.paths)))
        ]
        if config.logging.enabled:
            for line in summary:
                log.info(line.rstrip())
        else:
            return ''.join(summary)

    def save(self, filename, format=None, mode='nodes', weight=None):
        """Save the paths to file.

        Note
        ----
        Currently only the export to a pickle file or a ngram file is
        supported.

        Where pickles are a serialized byte stream of a Python object [1]_. This
        format will preserve Python objects used as nodes or edges.

        N-grams saves path data to a file containing multiple lines of n-grams
        of the form ``a,b,c,d,weight`` (where weight is optional). The default
        separating character ','. Each n-gram will be interpreted as a path.

        Parameters
        ----------
        filename : file or string
            File or filename to save. File ending such as '.pkl' is not
            necessary and will be added automatically if missing.

        format : string, optional (default = None)
            Format as which the network should be saved. Currently supported are
            '.pkl' files which are used by default, and 'ngram' files.

        mode : string, optional (default = 'nodes')
            Mode how the path should be saved. This option is not used when the
            paths are saved as a pickle file.

        weight : string or None, optional (default = None)
           The name of a path attribute that holds the numerical value used
           as a weight. If None or False, no weight will be assigned to the

        Examples
        --------
        >>> P = cn.Paths()
        >>> P.save('my_paths')

        References
        ----------
        .. [1] https://docs.python.org/3/library/pickle.html

        """
        # check paths
        if len(self.paths) > 0:
            # check mode
            if format is None:
                # check file ending and add .ngram if missing
                if not filename.endswith('.pkl'):
                    filename = filename + '.pkl'
                # write paths to output file
                with open(filename, 'wb') as f:
                    pickle.dump(deepcopy(self), f, pickle.HIGHEST_PROTOCOL)
            elif format == 'ngram':
                # check file ending and add .ngram if missing
                if not filename.endswith('.ngram'):
                    filename = filename + '.ngram'
                # write paths to output file
                with open(filename, 'w') as f:
                    for path in self.paths:
                        if weight is not None:
                            w = ','+str(path.weight(weight))
                        else:
                            w = ''
                        if mode == 'edges':
                            f.write(",".join(path.path.edges)+w+'\n')
                        elif mode == 'nodes':
                            f.write(",".join(path.path.nodes)+w+'\n')
                        else:
                            log.error("No mode '{}' is defined, hence no file"
                                      " was created.".format(format))

            else:
                log.warn("No format '{}' is defined, hence no file was"
                         "created.".format(path_format))

        else:
            log.warn("No 'paths' are defined, hence no file was created.")

    @classmethod
    def load(cls, filename, format=None):
        """Load paths from a file.

        Note
        ----
        Currently only the import from a pickle file is supported. Where pickles
        are a serialized byte stream of a Python object [1]_. This format will
        preserve Python objects used as nodes or edges.

        Parameters
        ----------
        filename : file or string
            File or filename to load. File ending such as '.pkl' is not
            necessary and will be added automatically if missing.

        format : string, optional (default = None)
            Format of the paths file.

        Returns
        -------
        paths : Paths
            Returns a cnet path.

        Examples
        --------
        >>> P_1 = cn.Paths()
        >>> P_1.save('my_paths')
        >>> P_2 = Network.load('my_paths')

        References
        ----------
        .. [1] https://docs.python.org/3/library/pickle.html

        """
        if format is None:
            # check file ending and add .pkl if missing
            if not filename.endswith('.pkl'):
                filename = filename + '.pkl'
            # write paths to output file
            with open(filename, 'rb') as f:
                paths = pickle.load(f)
        else:
            log.error('The format "{}" is currently not supported!'.format(format))
            raise NotImplementedError
        if isinstance(paths, Paths):
            return paths
        else:
            log.error(
                'The file "{}" does not contain a Network object'.format(filename))
            raise AttributeError

    def intersection(self, other):
        """Returns the paths which are common in both.

        Parameters
        ----------
        other : :py:class:`Paths`
            An Paths object storing paths.

        Returns
        -------
        paths : :py:class:`Paths`
            Returns a :py:class:`Paths` object with all parts common in this and
            the other object.

        Examples
        --------
        Create a list of paths.

        >>> p_1 = Path(['a', 'b'])
        >>> p_2 = Path(['a', 'b', 'c'])
        >>> P_1 = Paths([p_1])
        >>> P_2 = Paths([p_1, p_2])

        Get the interaction between P_1 and P_2.

        >>> P = P_1.interaction(P_2)
        >>> for p in P:
        >>>    print(p.name)
        a-b
        """
        paths = Paths(name='interaction between {} and {}'.format(
            self.name, other.name))
        paths.add_paths_from(list(set(self.paths).intersection(other.paths)))
        return paths

    def st_paths(self, source, target, mode='nodes'):
        """Returns all paths from the source to the target.

        Parameters
        ----------
        source : node id, edge id, Node object or Edge object
            Starting node or edge of the path.

        target : node id or edge id
            Ending node or edge for the path.

        mode : string, 'nodes' or 'edges', optional (default = 'nodes')
            The mode defines how the st-path is defined, i.e. as a list of node
            ids or a list of edge ids.

        Returns
        -------
        st_paths : :py:class:`Paths`
            Returns a :py:class:`Paths` object with all parts starting at the
            source and ending at the target.

        Examples
        --------
        Create a list of paths.

        >>> p_1 = Path(['a', 'b'])
        >>> p_2 = Path(['a', 'b', 'c'])
        >>> p_3 = Path(['a', 'c', 'b'])
        >>> P = Paths()
        >>> P.add_paths_from([p_1, p_2, p_3])

        Get the paths from node 'a' to 'b'.

        >>> P_ab = P.st_paths('a', 'b')
        >>> for p in P_ab:
        >>>     print(p.name)
        a-b
        a-c-b

        Get the paths from edge 'a-b' to 'b-c'.

        >>> P_ac = P.st_paths('a-b', 'b-c', mode='edges')
        >>> for p in P_ac:
        >>>     print(p.name)
        a-b-c

        """
        # check the source
        if mode == 'nodes':
            if isinstance(source, Node):
                _s = source.id
            else:
                _s = source
            if isinstance(target, Node):
                _t = target.id
            else:
                _t = target
        elif mode == 'edges':
            if isinstance(source, Edge):
                _s = source.id
            else:
                _s = source
            if isinstance(target, Edge):
                _t = target.id
            else:
                _t = target
        else:
            log.error('The mode "{}" is not supported!'.format(mode))
            raise NotImplementedError

        # create a new paths object
        st_paths = Paths(name='st-paths between {} and {}'.format(_s, _t))

        # go through the paths and add correct paths to the list
        if mode == 'nodes':
            for p in self.paths:
                if p.path[0] == _s and p.path[-1] == _t:
                    st_paths.add_path(p)

        if mode == 'edges':
            for p in self.paths:
                try:
                    _s_node = p.edge_to_nodes_map()[_s][0]
                    _t_node = p.edge_to_nodes_map()[_t][1]
                except:
                    _s_node = None
                    _t_node = None
                if p.path[0] == _s_node and p.path[-1] == _t_node:
                    st_paths.add_path(p)

        # return the st paths
        return st_paths

    def sort(self, key=None, reverse=False):
        """Sorts the paths in a specific order.

        Parameters
        ----------
        key : string, optional (default = None)
            Keyword that serves as a key for the sort comparison. Per default
            the list will be sorted by the length of the paths.

        reverse : Boole, optional (default = False)
            If true, the sorted list is reversed (or sorted in Descending order)

        Examples
        --------
        Create a list of paths.

        >>> p_1 = Path(['a', 'b', 'c'], cost=10)
        >>> p_2 = Path(['a', 'b'], cost=100)
        >>> p_3 = Path(['a', 'c', 'b'] cost=50)
        >>> P = Paths()
        >>> P.add_paths_from([p_1, p_2, p_3])

        Sort path by length

        >>> P.sort()
        >>> for p in P:
        >>>     print(p.name)
        a-b
        a-b-c
        a-c-b

        Sort by key value

        >>> P.sort('cost')
        >>> for p in P:
        >>>     print(p.name)
        a-b-c
        a-c-b
        a-b

        Reverse sort by key value

        >>> P.sort('cost', reverse=True)
        >>> for p in P:
        >>>     print(p.name)
        a-b
        a-c-b
        a-b-c

        """
        _list = []
        if key is None:
            for path in self.paths:
                _list.append((len(path), path))
        elif isinstance(key, str):
            for path in self.paths:
                try:
                    value = path[key]
                except:
                    log.error('The key {} is not defined for path {}'
                              ''.format(key, path.name))
                    raise
                _list.append((path[key], path))

        _list.sort(key=lambda x: x[0], reverse=reverse)
        for i, (value, path) in enumerate(_list):
            self.paths[i] = path

# =============================================================================
# eof
#
# Local Variables:
# mode: python
# mode: linum
# mode: auto-fill
# fill-column: 80
# End:



#!/usr/bin/python -tt
# -*- coding: utf-8 -*-
# =============================================================================
# File      : unit.py
# Creation  : 08 May 2018
# Time-stamp: <Don 2018-07-26 16:37 juergen>
#
# Copyright (c) 2018 Jürgen Hackl <hackl@ibi.baug.ethz.ch>
#               http://www.ibi.ethz.ch
# $Id$
#
# Description : Module to convert measurement units
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# =============================================================================

from . import logger
from .exceptions import CnetNotImplemented, CnetError
log = logger(__name__)


class UnitConverter(object):
    """Convert units.

    Parameters
    ----------
    input_unit : str, optional (default = 'cm')
        Unit which should be converted. The abbreviation of the unit is entered
        as string value. Currently supported are: Pixel 'px', Points 'pt',
        Millimeters 'mm', and Centimeters 'cm'.

    output_unit : str, optional (default = 'cm')
        Unit to which should be converted. The abbreviation of the unit is
        entered as string value. Currently supported are: Pixel 'px', Points
        'pt', Millimeters 'mm', and Centimeters 'cm'.

    digits : int, optional (default = 4)
        Number of digits to round the returning measure. Per default the
        measures are rounded to 4 digits.

    Examples
    --------
    >>> mm2cm = cn.UnitConverter('mm','cm')
    >>> mm2cm(10)
    1

    """

    def __init__(self, input_unit='cm', output_unit='cm', digits=4):
        """Initialize the unit converter.

        Parameters
        ----------
        input_unit : str, optional (default = 'cm')
            Unit which should be converted. The abbreviation of the unit is
            entered as string value. Currently supported are: Pixel 'px', Points
            'pt', Millimeters 'mm', and Centimeters 'cm'.

        output_unit : str, optional (default = 'cm')
            Unit to which should be converted. The abbreviation of the unit is
            entered as string value. Currently supported are: Pixel 'px', Points
            'pt', Millimeters 'mm', and Centimeters 'cm'.

        digits : int, optional (default = 4)
            Number of digits to round the returning measure. Per default the
            measures are rounded to 4 digits.
        """
        self.input_unit = input_unit
        self.output_unit = output_unit
        self.digits = digits

    def __call__(self, value):
        """Returns the converted measure.

        Returns
        -------
        measure : float
            Returns the converted measure.

        Examples
        --------
        >>> mm2cm = cn.UnitConverter('mm','cm')
        >>> mm2cm(10)
        1

        """
        return self.convert(value)

    @staticmethod
    def px_to_mm(measure):
        """Convert pixel to millimeters."""
        return measure * 0.26458333333719

    @staticmethod
    def px_to_pt(measure):
        """Convert pixel to points."""
        return measure * 0.75

    @staticmethod
    def pt_to_mm(measure):
        """Convert points to millimeters."""
        return measure * 0.352778

    @staticmethod
    def mm_to_px(measure):
        """Convert millimeters to pixel."""
        return measure * 3.779527559

    @staticmethod
    def mm_to_pt(measure):
        """Convert millimeters to points."""
        return measure * 2.83465

    def convert(self, value):
        """Returns the converted measure.

        Returns
        -------
        measure : float
            Returns the converted measure.

        Examples
        --------
        >>> mm2cm = cn.UnitConverter('mm','cm')
        >>> mm2cm.convert(10)
        1

        """
        try:
            measure = float(value)
        except:
            log.error('Value "{}" is not a number, and therefor can not'
                      ' converted to an other unit!.'.format(value))
            raise CnetError

        # to cm
        if self.input_unit == 'mm' and self.output_unit == 'cm':
            value = measure/10
        elif self.input_unit == 'pt' and self.output_unit == 'cm':
            value = self.pt_to_mm(measure)/10
        elif self.input_unit == 'px' and self.output_unit == 'cm':
            value = self.px_to_mm(measure)/10
        elif self.input_unit == 'cm' and self.output_unit == 'cm':
            value = measure
        # to pt
        elif self.input_unit == 'px' and self.output_unit == 'pt':
            value = self.px_to_pt(measure)
        elif self.input_unit == 'mm' and self.output_unit == 'pt':
            value = self.mm_to_pt(measure)
        elif self.input_unit == 'cm' and self.output_unit == 'pt':
            value = self.mm_to_pt(measure)*10
        elif self.input_unit == 'pt' and self.output_unit == 'pt':
            value = measure
        # to px
        elif self.input_unit == 'mm' and self.output_unit == 'px':
            value = self.mm_to_px(measure)
        elif self.input_unit == 'cm' and self.output_unit == 'px':
            value = self.mm_to_px(10*measure)
        elif self.input_unit == 'pt' and self.output_unit == 'px':
            value = measure*4/3
        elif self.input_unit == 'px' and self.output_unit == 'px':
            value = measure
        else:
            log.error('The conversion from "{}" to "{}" is currently not '
                      'supported!'.format(self.input_unit,
                                          self.output_unit))
            raise CnetNotImplemented

        # return the converted measure
        return round(value, self.digits)

# =============================================================================
# eof
#
# Local Variables:
# mode: python
# mode: linum
# mode: auto-fill
# fill-column: 80
# End:



from functools import singledispatch
from collections import defaultdict
import operator

import numpy as _np
import scipy.sparse as sparse
import scipy.sparse.linalg as sla
import scipy.linalg as la
import scipy as sp

from pathpy.utils import Log, Severity
from pathpy import HigherOrderNetwork
from pathpy import Network
from pathpy import Paths
from pathpy.algorithms.shortest_paths import *

from pathpy.utils import PathpyNotImplemented


__all__ = ["rank_centralities", "closeness_centrality", "betweenness_centrality",
           "eigenvector_centrality", "pagerank", 'node_traversals',
           'visitation_probabilities']


def rank_centralities(centralities):
    """Returns a list of (node, centrality) tuples in which tuples are ordered 
    by centrality in descending order

    Parameters
    ----------
    centralities: dict
        dictionary of centralities

    Examples
    --------
    >>> centralities = {'a': .2, 'b': .8, 'c': .5}
    >>> rank_centralities(centralities)
    [('b', 0.8), ('c', 0.5), ('a', 0.2)]

    Returns
    -------
    list 
        list of (node,centrality) tuples

    """
    ranked_nodes = sorted(centralities.items(), key=operator.itemgetter(1))
    ranked_nodes.reverse()
    return ranked_nodes



@singledispatch
def betweenness_centrality(network, normalized=False):
    assert isinstance(network, Network), \
        "network must be an instance of Network"

    Log.add('Calculating betweenness centralities in network ...', Severity.INFO)

    all_paths = shortest_paths(network)
    node_centralities = defaultdict(lambda: 0)

    for s in all_paths:
        for d in all_paths[s]:
            for p in all_paths[s][d]:
                for x in p[1:-1]:
                    if s != d != x:
                        node_centralities[x] += 1.0 / len(all_paths[s][d])
    if normalized:
        max_centr = max(node_centralities.values())
        for v in node_centralities:
            node_centralities[v] /= max_centr

    # assign zero values to nodes not occurring on shortest paths
    for v in network.nodes:
        node_centralities[v] += 0

    return node_centralities
    


@betweenness_centrality.register(HigherOrderNetwork)
def _bw(higher_order_net, normalized=False):
    """Calculates the betweenness centralities of all nodes.

    If the order of the higher-order network is larger than one
    centralities calculated based on the higher-order
    topology will automatically be projected back to first-order
    nodes.

    Parameters
    ----------
    network: HigherOrderNetwork
        an instance of a pathpy HigherOrderNetwork
    normalized:
        If set to True, betweenness centralities of nodes will be scaled by the maximum
        value (default False)

    Returns
    -------
    dict
        Dictionary containing as the keys the higher order node and as values their
        centralities
    """
    assert isinstance(higher_order_net, HigherOrderNetwork), \
        "arguments must be an instance of HigherOrderNetwork"

    Log.add('Calculating betweenness centralities (k = %s) ...' % \
        higher_order_net.order, Severity.INFO)

    all_paths = shortest_paths(higher_order_net)
    node_centralities = defaultdict(lambda: 0)

    shortest_paths_first_order = defaultdict(lambda: defaultdict(set))
    shortest_paths_first_order_lengths = defaultdict(lambda: defaultdict(lambda: _np.inf))

    for path_1_ord_k in all_paths:
        for path_2_ord_k in all_paths:
            source_k1 = higher_order_net.higher_order_node_to_path(path_1_ord_k)[0]
            dest_k1 = higher_order_net.higher_order_node_to_path(path_2_ord_k)[-1]

            # we consider a path in a k-th order network
            # connecting first-order node s1 to d1
            for path_ord_k in all_paths[path_1_ord_k][path_2_ord_k]:
                # convert k-th order path to first-order path and add
                #shortest_paths_first_order[source_k1][dest_k1].add(
                #    higher_order_net.higher_order_path_to_first_order(path_ord_k))

                p1 = higher_order_net.higher_order_path_to_first_order(path_ord_k)
                # obtain start node and end node
                s1 = p1[0]
                d1 = p1[-1]
                # compute the length of the first-order path
                l = len(p1) - 1
                # if path is a shortest path add it to dictionary
                if l < shortest_paths_first_order_lengths[s1][d1]:
                    shortest_paths_first_order_lengths[s1][d1] = l
                    shortest_paths_first_order[s1][d1] = set()
                    shortest_paths_first_order[s1][d1].add(p1)
                elif l == shortest_paths_first_order_lengths[s1][d1]:
                    shortest_paths_first_order[s1][d1].add(p1)

    for source_k1 in shortest_paths_first_order:
        for dest_k1 in shortest_paths_first_order[source_k1]:
            for path_k1 in shortest_paths_first_order[source_k1][dest_k1]:
                # increase betweenness centrality of all intermediary nodes
                # on path from s1 to d1
                for v in path_k1[1:-1]:
                    if source_k1 != v != dest_k1:
                        l_p = len(shortest_paths_first_order[source_k1][dest_k1])
                        node_centralities[v] += 1.0 / l_p
    if normalized:
        max_centr = max(node_centralities.values())
        for v in node_centralities:
            node_centralities[v] /= max_centr

    # assign centrality zero to nodes not occurring on higher-order shortest paths
    nodes = higher_order_net.paths.nodes
    for v in nodes:
        node_centralities[v] += 0

    Log.add('finished.', Severity.INFO)

    return node_centralities



@betweenness_centrality.register(Paths)
def _bw(paths, normalized=False):
    """Calculates the betweenness centrality of nodes based on observed shortest paths
    between all pairs of nodes

    Parameters
    ----------
    paths:
        Paths object
    normalized: bool
        normalize such that largest value is 1.0

    Returns
    -------
    dict
    """
    assert isinstance(paths, Paths), "argument must be an instance of pathpy.Paths"
    node_centralities = defaultdict(lambda: 0)

    Log.add('Calculating betweenness centralities in paths ...', Severity.INFO)

    all_paths = shortest_paths(paths)

    for s in all_paths:
        for d in all_paths[s]:
            for p in all_paths[s][d]:
                for x in p[1:-1]:
                    if s != d != x:
                        node_centralities[x] += 1.0 / len(all_paths[s][d])
    if normalized:
        max_centr = max(node_centralities.values())
        for v in node_centralities:
            node_centralities[v] /= max_centr

    # assign zero values to nodes not occurring on shortest paths
    nodes = paths.nodes
    for v in nodes:
        node_centralities[v] += 0
    Log.add('finished.')
    return node_centralities



@singledispatch
def closeness_centrality(network, normalized=False):
    """Calculates the closeness centralities of all nodes.

    If the order of the higher-order network is larger than one
    centralities calculated based on the higher-order
    topology will automatically be projected back to first-order
    nodes.

    Parameters
    ----------
    network: HigherOrderNetwork

    Returns
    -------
    dict
    """
    if not isinstance(network, Network):
        raise PathpyNotImplemented("`network` must be an instance of Network")

    distances = distance_matrix(network)
    node_centralities = defaultdict(lambda: 0)

    Log.add('Calculating closeness centralities in network ...', Severity.INFO)

    # calculate closeness values
    for x in network.nodes:
        for d in network.nodes:
            if d != x and distances[d][x] < _np.inf:
                node_centralities[x] += 1.0 / distances[d][x]

    # assign centrality zero to nodes not occurring on higher-order shortest paths
    for v in network.nodes:
        node_centralities[v] += 0

    if normalized:
        max_centr = max(node_centralities.values())
        for v in network.nodes:
            node_centralities[v] /= max_centr

    Log.add('finished.', Severity.INFO)

    return node_centralities



@closeness_centrality.register(Paths)
def _cl(paths, normalized=False):
    """Calculates the closeness centrality of nodes based on observed shortest paths
    between all nodes

    Parameters
    ----------
    paths: Paths
    normalized: bool
        normalize such that largest value is 1.0

    Returns
    -------
    dict
    """
    node_centralities = defaultdict(lambda: 0)
    distances = distance_matrix(paths)
    nodes = paths.nodes

    for x in nodes:
        # calculate closeness centrality of x
        for d in nodes:
            if x != d and distances[d][x] < _np.inf:
                node_centralities[x] += 1.0 / distances[d][x]

    # assign zero values to nodes not occurring
    
    for v in nodes:
        node_centralities[v] += 0

    if normalized:
        m = max(node_centralities.values())
        for v in nodes:
            node_centralities[v] /= m

    return node_centralities



@closeness_centrality.register(HigherOrderNetwork)
def _cl(higher_order_net, normalized=False):

    if not isinstance(higher_order_net, HigherOrderNetwork):
        raise PathpyNotImplemented("`higher_order_net` must be an instance of HigherOrderNetwork")

    distances = distance_matrix(higher_order_net)
    node_centralities = defaultdict(lambda: 0)
    nodes = higher_order_net.paths.nodes

    Log.add('Calculating closeness centralities (k = %s) ...' % higher_order_net.order,
            Severity.INFO)
    
    for x in nodes:
        # calculate closeness centrality of x
        for d in nodes:
            if x != d and distances[d][x] < _np.inf:
                node_centralities[x] += 1.0 / distances[d][x]

    # assign centrality zero to those nodes for which no higher-order path exists
    for v in nodes:
        node_centralities[v] += 0

    if normalized:
        m = max(node_centralities.values())
        for v in nodes:
            node_centralities[v] /= m

    Log.add('finished.', Severity.INFO)

    return node_centralities



def node_traversals(paths):
    """Calculates the number of times any path traverses each of the nodes.

    Parameters
    ----------
    paths: Paths

    Returns
    -------
    dict
    """
    if not isinstance(paths, Paths):
        raise PathpyNotImplemented("`paths` must be an instance of Paths")

    Log.add('Calculating node traversals...', Severity.INFO)

    # entries capture the number of times nodes are "visited by paths"
    # Note: this is identical to the subpath count of zero-length paths
    traversals = defaultdict(lambda: 0)

    for p in paths.paths[0]:
        traversals[p[0]] += paths.paths[0][p].sum()

    Log.add('finished.', Severity.INFO)

    return traversals



def visitation_probabilities(paths):
    """Calculates the probabilities that a randomly chosen path passes through each of
    the nodes. If 5 out of 100 paths (of any length) traverse node v, node v will be
    assigned a visitation probability of 0.05. This measure can be interpreted as ground
    truth for the notion of importance captured by PageRank applied to a graphical
    abstraction of the paths.

    Parameters
    ----------
    paths: Paths

    Returns
    -------
    dict
    """
    if not isinstance(paths, Paths):
        raise PathpyNotImplemented("`paths` must be an instance of Paths")
    Log.add('Calculating visitation probabilities...', Severity.INFO)

    # entries capture the probability that a given node is visited on an arbitrary path
    # Note: this is identical to the subpath count of zero-length paths
    # (i.e. the relative frequencies of nodes across all pathways)
    visit_probabilities = node_traversals(paths)

    # total number of visits
    visits = 0.0
    for v in visit_probabilities:
        visits += visit_probabilities[v]

    for v in visit_probabilities:
        visit_probabilities[v] /= visits

    Log.add('finished.', Severity.INFO)

    return visit_probabilities



def eigenvector_centrality(network, projection='scaled', include_sub_paths=True):
    """Calculates the eigenvector centralities of higher-order nodes.

    If the order of the HigherOrderNetwork is larger than one, the centralities
    will be projected to the first-order nodes.

    Parameters
    ----------
    network
    projection: str
        Indicates how the projection from k-th-order nodes (v1, v2, ... , v{k-1}) shall be
        performed. For the method 'all', the eigenvector centrality of the higher-order
        node will be added to *all* first-order nodes on the path corresponding to the
        higher-order node. For the method 'last', the centrality of the higher-order node
        will only be assigned to *last* first-order node v{k-1}. For the method 'scaled'
        (default), the eigenvector centrality of higher-order nodes will be assigned
        proportionally to first-order nodes, i.e. each of the three nodes in the
        third-order node (a,b,c) will receive one third of the eigenvector centrality
        of (a,b,c).
    include_sub_paths: bool
        whether or not to include subpath statistics in the calculation (default True)

    Returns
    -------
    dict

    """
    assert isinstance(network, HigherOrderNetwork), \
        "network must be an instance of HigherOrderNetwork"
    adj_mat = network.adjacency_matrix(include_subpaths=include_sub_paths,
                                       weighted=False, transposed=True)

    # calculate leading eigenvector of A
    _, v = sla.eigs(adj_mat, k=1, which="LM", ncv=13)

    v = v.reshape(v.size, )

    higher_order_eigen_vec_cent = dict(zip(network.nodes, map(_np.abs, v)))

    # project eigen_vec_cent of higher-order nodes to first-order network
    first_order_eigen_vec_cent = defaultdict(lambda: 0.0)

    # sum eigen_vec_cent values based on higher-order nodes
    # and normalize the result
    for v in network.nodes:
        # turns node a-b-c in path tuple (a,b,c)
        p = network.higher_order_node_to_path(v)
        if projection == 'all':
            # assign eigen_vec_cent of higher-order node to all first-order nodes
            for x in p:
                first_order_eigen_vec_cent[x] += higher_order_eigen_vec_cent[v]
        elif projection == 'scaled':
            for x in p:
                n_p = float(len(p))
                first_order_eigen_vec_cent[x] += higher_order_eigen_vec_cent[v] / n_p
        elif projection == 'last':
            # assign eigen_vec_cent  of higher-order node to last first-order node
            first_order_eigen_vec_cent[p[-1]] += higher_order_eigen_vec_cent[v]
        elif projection == 'first':
            # assign eigen_vec_cent  of higher-order node to last first-order node
            first_order_eigen_vec_cent[p[0]] += higher_order_eigen_vec_cent[v]

    # for scaled, values sum to one anyway
    if projection != 'scaled':
        for v in first_order_eigen_vec_cent:
            first_order_eigen_vec_cent[v] /= sum(first_order_eigen_vec_cent.values())

    Log.add('finished.', Severity.INFO)

    return first_order_eigen_vec_cent



def pagerank(network, alpha=0.85, max_iter=100, tol=1.0e-6, projection='scaled',
             include_sub_paths=True, weighted=False):
    """Calculates the PageRank of higher-order nodes based on a power iteration.

    If the order of the higher-order network is larger than one, the PageRank calculated
    based on the higher-order topology will automatically be projected back to first-order
    nodes.

    Parameters
    ----------
    network: HigherOrderNetwork
    alpha: float
        damping factor
    max_iter: int
        maximum number or iterations in solver
    tol: float
        accepted tolerance for convergence check
    projection: str
        Indicates how the projection from k-th-order nodes (v1, v2, ... , v{k-1}) shall be
        performed. For the method 'all', the PageRank value of the higher-order node will
        be added to *all* first-order nodes on the path corresponding to the higher-order
        node. For the method 'last', the PR value of the higher-order node will only be
        assigned to *last* first-order node v{k-1}. For the method 'scaled' (default), the
        PageRank of higher-order nodes will be assigned proportionally to first-order
        nodes, i.e. each of the three nodes in the third-order node (a,b,c) will receive
        one third of the PageRank of (a,b,c).
    include_sub_paths: bool
        whether or not to use subpath statistics in the PageRank calculation
    weighted: bool
        use path weights in the calculation

    Returns
    -------
    dict

    """
    assert isinstance(network, HigherOrderNetwork), \
        "network must be an instance of HigherOrderNetwork"
    assert projection in ['all', 'last', 'first', 'scaled'], 'Invalid projection method'

    Log.add('Calculating PageRank in ' + str(network.order) + '-th order network...',
            Severity.INFO)

    higher_order_pr = defaultdict(lambda: 0)

    n_nodes = float(len(network.nodes))

    assert n_nodes > 0, "Number of nodes is zero"

    # entries A[s,t] give directed link s -> t
    adj_mat = network.adjacency_matrix(include_subpaths=include_sub_paths,
                                       weighted=weighted, transposed=False)

    # sum of outgoing node degrees
    row_sums = sp.array(adj_mat.sum(axis=1)).flatten()

    # replace non-zero entries x by 1/x
    row_sums[row_sums != 0] = 1.0 / row_sums[row_sums != 0]

    # indices of zero entries in row_sums
    d = sp.where(row_sums == 0)[0]

    # create sparse matrix with row_sums as diagonal elements
    q_mat = sparse.spdiags(row_sums.T, 0, adj_mat.shape[0], adj_mat.shape[1],
                           format='csr')

    # with this, we have divided elements in non-zero rows in A by 1 over the row sum
    q_mat = q_mat * adj_mat

    # vector with n entries 1/n
    inv_n_nodes = sp.array([1.0 / n_nodes] * int(n_nodes))

    p_rank = inv_n_nodes

    # Power iteration
    for _ in range(max_iter):
        last = p_rank

        # sum(pr[d]) is the sum of PageRanks for nodes with zero out-degree
        # sum(pr[d]) * p yields a vector with length n
        p_rank = (alpha * (p_rank * q_mat + sum(p_rank[d]) * inv_n_nodes) +
                  (1 - alpha) * inv_n_nodes)

        if sp.absolute(p_rank - last).sum() < n_nodes * tol:
            higher_order_pr = dict(zip(network.nodes, map(float, p_rank)))
            break

    if network.order == 1:
        return higher_order_pr

    # project PageRank of higher-order nodes to first-order network
    first_order_pr = defaultdict(lambda: 0.0)

    # sum PageRank values based on higher-order nodes
    # and normalize the result
    for v in network.nodes:
        # turns node a-b-c in path tuple (a,b,c)
        inv_n_nodes = network.higher_order_node_to_path(v)
        if projection == 'all':
            # assign PR of higher-order node to all first-order nodes
            for x in inv_n_nodes:
                first_order_pr[x] += higher_order_pr[v] / len(inv_n_nodes)
        elif projection == 'scaled':
            for x in inv_n_nodes:
                # each node on e.g. a 4-th-order path a-b-c-d receives one fourth of the
                # PageRank value, to ensure that the resulting first-order PageRank sums
                # to one
                first_order_pr[x] += higher_order_pr[v] / float(len(inv_n_nodes))
        elif projection == 'last':
            # assign PR of higher-order node to last first-order node
            first_order_pr[inv_n_nodes[-1]] += higher_order_pr[v]
        elif projection == 'first':
            # assign PR of higher-order node to last first-order node
            first_order_pr[inv_n_nodes[0]] += higher_order_pr[v]

    # for projection method 'scaled', the values sum to one anyway
    if projection != 'scaled':
        for v in first_order_pr:
            first_order_pr[v] /= sum(first_order_pr.values())

    # assign centrality zero to nodes not occurring in higher-order PR
    nodes = network.paths.nodes
    for v in nodes:
        first_order_pr[v] += 0

    Log.add('finished.', Severity.INFO)

    return first_order_pr



# -*- coding: utf-8 -*-

#    pathpy is an OpenSource python package for the analysis of time series data
#    on networks using higher- and multi order graphical models.
#
#    Copyright (C) 2016-2017 Ingo Scholtes, ETH Zürich
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    Contact the developer:

#    E-mail: ischoltes@ethz.ch
#    Web:    http://www.ingoscholtes.net
from collections import defaultdict

import numpy as np
import scipy.sparse.linalg as sla

from pathpy import HigherOrderNetwork
from pathpy.utils import Log, Severity
from pathpy import Paths
from pathpy.utils import PathpyError


__all__ = ['slow_down_factor', 'entropy_growth_rate_ratio',
           'betweenness_preference', 'betweenness_preference_matrix']


def slow_down_factor(paths, k=2, lanczos_vectors=15, maxiter=1000):
    """Returns a factor S that indicates how much slower (S>1) or faster (S<1)
    a diffusion process evolves in a k-order model of the path statistics
    compared to what is expected based on a first-order model. This value captures
    the effect of order correlations of length k on a diffusion process which evolves
    based on the observed paths.

    Parameters
    ----------
    paths: Paths
        the Paths object to compute the slow down factor for
    k: int
        order of the model to consider
    lanczos_vectors: int
        number of Lanczos vectors to be used in the approximate calculation of
        eigenvectors and eigenvalues. This maps to the ncv parameter of scipy's
        underlying function eigs.
    maxiter: int
        scaling factor for the number of iterations to be used in the approximate
        calculation of eigenvectors and eigenvalues. The number of iterations passed to
        scipy's underlying eigs function will be n*maxiter where n is the number of
        rows/columns of the Laplacian matrix.

    Returns
    -------
    float
    """
    assert k > 1, 'Slow-down factor can only be calculated for orders larger than one'

    # NOTE to myself: most of the time goes for construction of the 2nd order
    # NOTE            null graph, then for the 2nd order null transition matrix

    gk = HigherOrderNetwork(paths, k=k)
    gkn = HigherOrderNetwork(paths, k=k, null_model=True)

    Log.add('Calculating slow down factor ... ', Severity.INFO)

    # Build transition matrices
    Tk = gk.transition_matrix()
    Tkn = gkn.transition_matrix()

    # Compute eigenvector sequences
    # NOTE: ncv=13 sets additional auxiliary eigenvectors that are computed
    # NOTE: in order to be more confident to find the one with the largest
    # NOTE: magnitude, see
    # NOTE: https://github.com/scipy/scipy/issues/4987
    w2 = sla.eigs(Tk, which="LM", k=2, ncv=lanczos_vectors, return_eigenvectors=False,
                  maxiter=maxiter)
    eigen_values2_sorted = np.sort(-np.absolute(w2))

    w2n = sla.eigs(Tkn, which="LM", k=2, ncv=lanczos_vectors, return_eigenvectors=False,
                   maxiter=maxiter)
    eigen_values2n_sorted = np.sort(-np.absolute(w2n))

    Log.add('finished.', Severity.INFO)

    abs_eigen_values2 = np.abs(eigen_values2_sorted[1])
    abs_eigen_values2n = np.abs(eigen_values2n_sorted[1])
    return np.log(abs_eigen_values2n) / np.log(abs_eigen_values2)


def entropy_growth_rate_ratio(paths, method='MLE', k=2, lanczos_vectors=15, maxiter=1000):
    """Computes the ratio between the entropy growth rate ratio between
    the k-order and first-order model of a temporal network t. Ratios smaller
    than one indicate that the temporal network exhibits non-Markovian characteristics

    Parameters
    ----------
    paths: Paths
    method:
        which method to use for the entropy calculation. The default 'MLE' uses
        the standard Maximum-Likelihood estimation of entropy. Setting method to
        'Miller' additionally applies a Miller-correction.
        see e.g.
    k: int
        order of the higher order network to generate
    lanczos_vectors:
        number of Lanczos vectors to be used in the approximate calculation of
        eigenvectors and eigenvalues. This maps to the ncv parameter of scipy's underlying
        function eigs.
    maxiter:
        Maximum number of Arnoldi update iterations allowed Default: n*10
        See scipy.sparse.linalg.eigs for more details


    References
    ---------
        Liam Paninski: Estimation of Entropy and Mutual Information, Neural Computation 5,
        2003 or
        http://www.nowozin.net/sebastian/blog/estimating-discrete-entropy-part-2.html

    Returns
    -------
    float
    """
    # NOTE to myself: most of the time here goes into computation of the
    # NOTE            EV of the transition matrix for the bigger of the
    # NOTE            two graphs below (either 2nd-order or 2nd-order null)

    assert (method == 'MLE' or method == 'Miller'), \
        'Only methods MLE or Miller are supported'

    # Generate k-order network
    gk = HigherOrderNetwork(paths, k=k)
    g1 = HigherOrderNetwork(paths, k=1)

    Log.add('Calculating entropy growth rate ratio ... ', Severity.INFO)

    # Compute entropy growth rate of observed transition matrix
    adj_matrix = g1.adjacency_matrix(weighted=False, transposed=True)
    tran_mat_k = gk.transition_matrix()
    leading_eigen_vec_k = HigherOrderNetwork.leading_eigenvector(
        tran_mat_k, normalized=True, lanczos_vecs=lanczos_vectors, maxiter=maxiter
    )

    tran_mat_k.data *= np.log2(tran_mat_k.data)

    # Apply Miller correction to the entropy estimation
    if method == 'Miller':
        # Here, 'possible_k_paths' is the number of different k-paths that can exist
        # based on the observed edges
        possible_k_paths = (adj_matrix ** k).sum()
        print('K = ', possible_k_paths)

        # N is the number of observations used to estimate the transition probabilities
        # in the second-order network. This corresponds to the total edge weight in the
        # k-order network, or - alternatively - to the number of paths of length k
        num_obs = 0
        for p in paths.paths[k]:
            num_obs += paths.paths[k][p].sum()
        print('N = ', num_obs)
        h_mat_k = (- np.sum(tran_mat_k * leading_eigen_vec_k) +
                   (possible_k_paths - 1) / (2 * num_obs))
    else:
        # simple MLE estimation
        h_mat_k = - np.sum(tran_mat_k * leading_eigen_vec_k)

    h_mat_k = np.absolute(h_mat_k)

    # Compute entropy rate of first-order model
    gk_n = HigherOrderNetwork(paths, k=1, null_model=False)

    # For the entropy rate of the null model, no Miller correction is needed
    # since we assume that transitions correspond to the true probabilities
    trans_mat_null = gk_n.transition_matrix()
    leading_eigen_v_null = HigherOrderNetwork.leading_eigenvector(trans_mat_null)
    trans_mat_null.data *= np.log2(trans_mat_null.data)
    h_mat_null = -np.sum(trans_mat_null * leading_eigen_v_null)
    h_mat_null = np.absolute(h_mat_null)

    Log.add('finished.', Severity.INFO)

    print('entropy rate (k) = ', h_mat_k)
    print('entropy rate (1) = ', h_mat_null)
    # Return ratio
    return h_mat_k / h_mat_null


def betweenness_preference_matrix(paths, v):
    """Computes a betweenness preference matrix for a node v

    Parameters
    ----------
    paths: Paths
    v:
        Node for which the betweenness preference matrix shall
        be calculated

    Returns
    -------

    """
    # create first-order network
    g = HigherOrderNetwork(paths)

    in_degree = len(g.predecessors[v])
    out_degree = len(g.successors[v])

    index_succ = {}
    index_pred = {}

    B_v = np.zeros(shape=(in_degree, out_degree))

    # Create an index-to-node mapping for predecessors and successors
    i = 0
    for u in g.predecessors[v]:
        index_pred[u] = i
        i = i + 1

    i = 0
    for w in g.successors[v]:
        index_succ[w] = i
        i = i + 1

    # Calculate entries of betweenness preference matrix
    for p in paths.paths[2]:
        if p[1] == v:
            B_v[index_pred[p[0]], index_succ[p[2]]] += paths.paths[2][p].sum()

    return B_v


def betweenness_preference(paths, v, normalized=False, method='MLE'):
    """Calculates the betweenness preferences of a node v based on the mutual information
     of path statistics of length two.

    Parameters
    ----------
    paths: Paths
    v:
        Node for which the betweenness preference will be calculated
    normalized: bool
        whether or not to normalize betweenness preference values
    method: str
        which method to use for the entropy calculation. The default 'MLE' uses the
        standard Maximum-Likelihood estimation of entropy. Setting method to 'Miller'
        additionally applies a Miller-correction.

    References
    ---------
        see e.g. Liam Paninski: Estimation of Entropy and Mutual Information,
        Neural Computation 5, 2003 or
        http://www.nowozin.net/sebastian/blog/estimating-discrete-entropy-part-2.html

    Returns
    -------

    """
    assert method in ['MLE', 'Miller'], "method must be one of 'MLE' or 'Miller'"

    # If the network is empty, just return zero
    if not paths.nodes:
        return 0.0

    # First create the betweenness preference matrix (equation (2) of the paper)
    B_v = betweenness_preference_matrix(paths, v)

    if B_v.shape[0] == 0 or B_v.shape[1] == 0:
        return None

    # Normalize matrix (equation (3) of the paper)
    # NOTE: P_v has the same shape as B_v
    P_v = np.zeros(shape=B_v.shape)
    S = np.sum(B_v)

    if S > 0:
        P_v = B_v / S

    # Compute marginal probabilities
    # Marginal probabilities P^v_d = \sum_s'{P_{s'd}}
    marginal_d = np.sum(P_v, axis=0)

    # Marginal probabilities P^v_s = \sum_d'{P_{sd'}}
    marginal_s = np.sum(P_v, axis=1)

    if method == 'Miller':

        # total number of samples, i.e. observed two-paths
        N = np.sum(B_v)

        # print('N = ', N)
        # print('B = ', B_v)
        # print('marginal_s = ', marginal_s)
        # print('marginal_d = ', marginal_d)

        # marginal entropy H(S)
        H_s = _entropy(marginal_s, len(marginal_s), N, method='Miller')

        # print('H(S) = ', H_s)
        # marginal entropy H(D)

        H_d = _entropy(marginal_d, len(marginal_d), N, method='Miller')

        # print('H(D) = ', H_d)
        # we need the conditional entropy H(D|S)

        H_ds = 0
        for s in range(len(marginal_s)):
            # number of two paths s -> v -> * observed in the data
            N_s = np.sum(B_v[s, :])

            # print('N(s=' + str(s) + ') = ' +  str(N_s))

            # probabilities of all destinations, given the particular source s
            p_ds = B_v[s, :] / np.sum(B_v[s, :])

            # print('P(D|S=' + str(s) + ') = '+ str(p_ds))

            # number of possible destinations d
            K_s = len(p_ds)

            # print('K(s=' + str(s) + ') = ' +  str(K_s))

            # marginal_s[s] is the overall probability of source s
            p_s = marginal_s[s]

            # add to conditional entropy
            H_ds += p_s * _entropy(p_ds, K_s, N_s, method='Miller')

            I = H_d - H_ds

        # print('H(D|S) = ', H_ds)

    else:
        # use MLE estimation
        H_s = _entropy(marginal_s)
        H_d = _entropy(marginal_d)
        # H_ds = 0

        # for s in range(len(marginal_s)):
        #    print('s = ' + str(s) + ': ' + str(np.sum(P_v[s,:])))
        #    p_ds = P_v[s,:]/np.sum(P_v[s,:])
        #    H_ds += marginal_s[s] * Paths.__Entropy(p_ds)

        # Alternative calculation (without explicit entropies)
        # build mask for non-zero elements
        row, col = np.nonzero(P_v)
        pv = P_v[(row, col)]
        marginal = np.outer(marginal_s, marginal_d)
        log_argument = np.divide(pv, marginal[(row, col)])
        I = np.dot(pv, np.log2(log_argument))

    # I = H_d - H_ds

    if normalized:
        I = I / np.min([H_s, H_d])

    return I


def _entropy(prob, possible_outcomes=None, sample_size=None, method='MLE'):
    """Calculates the entropy of an (observed) probability distribution based on Maximum
    Likelihood Estimation (MLE) (default) or using a Miller correction.

    Parameters
    ----------
    prob:
        numpy array of the observed probabilities
    possible_outcomes: int
        the number of possible outcomes, i.e. outcomes with non-zero probability to be
        used for the Miller correction (default None)
    sample_size: int
        number of samples based on which observed probabilities where computed. This is
        needed for the Miller correction (default None)
    method: str
        The method to be used to calculate entropy. Can be 'MLE' (default) or 'Miller'

    Returns
    -------

    """
    assert method in ["MLE", "Miller"], "method must be one of 'MLE' or 'Miller'"
    if method == 'MLE':
        idx = np.nonzero(prob)
        return -np.inner(np.log2(prob[idx]), prob[idx])
    elif method == 'Miller':
        assert (possible_outcomes is not None) and (sample_size is not None)
        if sample_size == 0:
            return 0

        idx = np.nonzero(prob)

        addition = (possible_outcomes - 1) / (2 * sample_size)
        return -np.inner(np.log2(prob[idx]), prob[idx]) + addition
    else:
        raise PathpyError("method {} not supported".format(method))



#!/usr/bin/python -tt
# -*- coding: utf-8 -*-
# =============================================================================
# File      : drawing.py
# Creation  : 08 May 2018
# Time-stamp: <Die 2018-08-07 10:40 juergen>
#
# Copyright (c) 2018 Jürgen Hackl <hackl@ibi.baug.ethz.ch>
#               http://www.ibi.ethz.ch
# $Id$
#
# Description : Module to draw the network
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# =============================================================================

import numpy as np
from collections import OrderedDict
from . import logger
from .exceptions import CnetError
from .units import UnitConverter
from .canvas import Canvas
from .layout import Layout
log = logger(__name__)

# TODO: move this to the config file
DIGITS = 3


class TikzNetworkDrawer(object):
    """Class which handles the drawing of the network.

    The default visualization of a network draws the nodes on a 2D plane
    according to a given Layout, then draws a straight or curved edge between
    nodes connected by edges. This is the visualization used when one invokes
    the :py:meth:`plot` function on a Network object.

    Parameters
    ----------
    network : network object
        Network to be drawn. The network can be a 'cnet', 'networkx', 'igraph',
        'pathpy' object, or a tuple of a node list and edge list.

    kwds : keyword arguments, optional (default= no attributes)
        Attributes to add to the drawer as key=value pairs.
        See also :py:meth:`plot`

    See Also
    --------
    plot

    """

    def __init__(self, network, **kwds):
        """Initialize the network drawer.

        Parameters
        ----------
        network : network object
            Network to be drawn. The network can be a 'cnet', 'networkx',
            'igraph', 'pathpy' object, or a tuple of a node list and edge list.

        kwds : keyword arguments, optional (default= no attributes)
            Attributes to add to the drawer as key=value pairs.
            See also :py:meth:`plot`

        """
        # initialize variables
        self.edges = OrderedDict()
        self.nodes = None
        self.directed = False
        self.digits = DIGITS

        self.node_attributes = {}
        self.edge_attributes = {}
        self.layout_attributes = {}
        self.general_attributes = {}

        self.adjacency_matrix = None

        # TODO: find a nicer way to do this checks
        # check if a layout is given or should be generated
        _layout = False
        if isinstance(kwds.get('layout', None), str):
            _layout = True
        # check if the layout is weighted
        _weight = kwds.get('weight', None)
        if _weight is None:
            _weight = kwds.get('layout_weight', None)

        # check type of network
        if 'cnet' in str(type(network)):
            # log.debug('The network is of type "cnet".')
            for e, n in network.edges(nodes=True):
                self.edges[e] = n
            self.nodes = list(network.nodes)
            self.directed = network.directed
            if _layout:
                self.adjacency_matrix = network.adjacency_matrix(weight=_weight)

        elif 'networkx' in str(type(network)):
            # log.debug('The network is of type "networkx".')
            for e in network.edges():
                self.edges[e] = e
            self.nodes = list(network.nodes())
            self.directed = network.is_directed()
            if _layout:
                import networkx as nx
                self.adjacency_matrix = nx.adjacency_matrix(
                    network, weight=_weight)

        elif 'igraph' in str(type(network)):
            # log.debug('The network is of type "igraph".')
            for i in range(len(network.es)):
                self.edges[i] = network.es[i].tuple
            self.nodes = list(range(len(network.vs)))
            self.directed = network.is_directed()
            if _layout:
                from scipy.sparse import coo_matrix
                A = np.array(network.get_adjacency(attribute=_weight).data)
                self.adjacency_matrix = coo_matrix(A)

        elif 'pathpy' in str(type(network)):
            # log.debug('The network is of type "pathpy".')
            for e in network.edges:
                self.edges[e] = e
            self.nodes = list(network.nodes)
            self.directed = network.directed
            if _layout:
                if _weight is not None:
                    _w = True
                else:
                    _w = False
            self.adjacency_matrix = network.adjacency_matrix(weighted=_w)

        elif isinstance(network, tuple):
            # log.debug('The network is of type "list".')
            self.nodes = network[0]
            for e in network[1]:
                self.edges[e] = e

        else:
            log.error('Type of the network could not be determined.'
                      ' Currently only "cnet", "networkx","igraph", "pathpy"'
                      ' and "node/edge list" is supported!')
            raise CnetNotImplemented

        # assign attributes to the class
        self.attributes = kwds
        # draw the network in the memory
        self.draw()

    @staticmethod
    def rename_attributes(**kwds):
        """Rename node and edge attributes.

        In the style dictionary multiple keywords can be used to address
        attributes. These keywords will be converted to an unique key word,
        used in the remaining code. This allows to keep the keywords used in
        'igrap'.

        ========= ===========================
        keys      other valid keys
        ========= ===========================
        node      vertex, v, n
        edge      link, l, e
        margins   margin
        canvas    bbox, figure_size
        units     unit
        ========= ===========================

        """
        names = {'node_': ['vertex_', 'v_', 'n_'],
                 'edge_': ['edge_', 'link_', 'l_', 'e_'],
                 'margins': ['margin'],
                 'canvas': ['bbox', 'figure_size'],
                 'units': ['units', 'unit'],
                 }
        _kwds = {}
        del_keys = []
        for key, value in kwds.items():
            for attr, name_list in names.items():
                for name in name_list:
                    if name in key and name[0] == key[0]:
                        _kwds[key.replace(name, attr)] = value
                        del_keys.append(key)
                        break
        # remove the replaced keys from the dict
        for key in del_keys:
            del kwds[key]

        return {**_kwds, **kwds}

    def draw(self):
        """Function to draw a virtual network."""
        # rename the attributes
        kwds = self.rename_attributes(**self.attributes)

        # go through all attributes and assign them to nodes, edges or the
        # general dictionary
        for key, value in kwds.items():
            if 'node_' in key:
                self.node_attributes[key] = self.format_node_value(value)
            elif 'edge_' in key:
                self.edge_attributes[key] = self.format_edge_value(value)
            # elif 'layout_' in key:
            #     if 'fixed' in key or 'positions' in key:
            #         self.layout_attributes[key] = self.format_node_value(value)
            #     elif 'weight' in key:
            #         self.layout_attributes[key] = self.format_edge_value(value)
            #     else:
            #         self.layout_attributes[key] = value
            else:
                if 'fixed' in key or 'positions' in key:
                    self.general_attributes[key] = self.format_node_value(value)
                elif 'weight' in key:
                    self.general_attributes[key] = self.format_edge_value(value)
                else:
                    self.general_attributes[key] = value

        # check if network is directed and nothing other is defined
        if self.directed and \
           self.edge_attributes.get('edge_directed', None) is None:
            self.edge_attributes['edge_directed'] = self.format_edge_value(True)

        # convert the units
        self.convert_units()

        # create canvas
        _canvas = self.general_attributes.get('canvas', (None, None))
        _margins = self.general_attributes.get('margins', None)
        _node_sizes = self.node_attributes.get('node_size', None)
        self.canvas = Canvas(_canvas[0], _canvas[1], _margins, _node_sizes)

        # configure the layout
        # check if a layout is defined
        _layout = self.attributes.get('layout', None)
        if isinstance(_layout, str) or _layout is None:
            self.general_attributes['layout'] = _layout
            self.layout = Layout(self.nodes, self.adjacency_matrix,
                                 **self.general_attributes).generate_layout()
        else:
            self.layout = self.format_node_value(self.attributes['layout'])

        # self.layout = {}
        # for node in self.nodes:
        #     self.layout[node] = (np.random.rand(), np.random.rand())

        # fit the node position to the chosen canvas
        k_a_r = self.general_attributes.get('keep_aspect_ratio', True)
        self.layout = self.canvas.fit(self.layout, keep_aspect_ratio=k_a_r)

        # assign layout to the nodes
        self.node_attributes['layout'] = self.layout

        # bend the edges if enabled
        if self.edge_attributes.get('edge_curved', None) is not None:
            self.edge_attributes['edge_curved'] = self.curve()

        # initialize vertices
        self.node_drawer = []
        for node in self.nodes:
            _attr = {}
            for key in self.node_attributes:
                _attr[key] = self.node_attributes[key][node]
            self.node_drawer.append(TikzNodeDrawer(node, **_attr))

        # initialize edges
        self.edge_drawer = []
        for edge, (u, v) in self.edges.items():
            _attr = {}
            for key in self.edge_attributes:
                _attr[key] = self.edge_attributes[key][edge]
            self.edge_drawer.append(TikzEdgeDrawer(edge, u, v, **_attr))

    def convert_units(self):
        """Function to convert the units used."""
        # get unit converter
        _units = self.general_attributes.get('units', ('cm', 'pt'))
        if isinstance(_units, tuple):
            self.unit2cm = UnitConverter(_units[0], 'cm')
            self.unit2pt = UnitConverter(_units[1], 'pt')
        else:
            self.unit2cm = UnitConverter(_units, 'cm')
            self.unit2pt = UnitConverter(_units, 'pt')

        # TODO: Fix this ugly code!
        for key in ['node_size', 'node_label_distance']:
            if key in self.node_attributes:
                _attr = {}
                for k, v in self.node_attributes[key].items():
                    if isinstance(v, int) or isinstance(v, float):
                        _attr[k] = self.unit2cm(v)
                    else:
                        _attr[k] = v
                self.node_attributes[key] = _attr

        for key in ['node_label_size']:
            if key in self.node_attributes:
                _attr = {}
                for k, v in self.node_attributes[key].items():
                    if isinstance(v, int) or isinstance(v, float):
                        _attr[k] = round(self.unit2pt(v)/7, self.digits)
                    else:
                        _attr[k] = v
                self.node_attributes[key] = _attr

        for key in ['edge_arrow_size', 'edge_arrow_width']:
            if key in self.edge_attributes:
                _attr = {}
                for k, v in self.edge_attributes[key].items():
                    if isinstance(v, int) or isinstance(v, float):
                        _attr[k] = self.unit2cm(v)
                    else:
                        _attr[k] = v
                self.edge_attributes[key] = _attr

        for key in ['edge_width']:
            if key in self.edge_attributes:
                _attr = {}
                for k, v in self.edge_attributes[key].items():
                    if isinstance(v, int) or isinstance(v, float):
                        _attr[k] = self.unit2pt(v)
                    else:
                        _attr[k] = v
                self.edge_attributes[key] = _attr

        for key in ['edge_loop_size']:
            if key in self.edge_attributes:
                _attr = {}
                for k, v in self.edge_attributes[key].items():
                    if isinstance(v, int) or isinstance(v, float):
                        _attr[k] = str(self.unit2cm(v))+'cm'
                    else:
                        _attr[k] = v
                self.edge_attributes[key] = _attr

        for key in ['edge_label_size']:
            if key in self.edge_attributes:
                _attr = {}
                for k, v in self.edge_attributes[key].items():
                    if isinstance(v, int) or isinstance(v, float):
                        _attr[k] = round(self.unit2pt(v)/7, self.digits)
                    else:
                        _attr[k] = v
                self.edge_attributes[key] = _attr

        if 'canvas' in self.general_attributes:
            w, h = self.general_attributes['canvas']
            self.general_attributes['canvas'] = (
                self.unit2cm(w), self.unit2cm(h))

        if 'margins' in self.general_attributes:
            _margins = self.general_attributes['margins']
            if isinstance(_margins, int) or isinstance(_margins, float):
                value = self.unit2cm(_margins)
            else:
                value = {'top': self.unit2cm(_margins.get('top', 0)),
                         'left': self.unit2cm(_margins.get('left', 0)),
                         'bottom': self.unit2cm(_margins.get('bottom', 0)),
                         'right': self.unit2cm(_margins.get('right', 0))}
            self.general_attributes['margins'] = value

        for key in ['xshift', 'yshift']:
            if key in self.general_attributes:
                v = self.general_attributes[key]
                self.general_attributes[key] = str(self.unit2cm(v))+'cm'

    def format_node_value(self, value):
        """Returns a dict with node ids and assigned values."""
        # check if value is string, list or dict
        _values = {}
        if isinstance(value, str) or isinstance(value, int) or \
           isinstance(value, float) or isinstance(value, tuple):
            for n in self.nodes:
                _values[n] = value
        elif isinstance(value, list):
            for i, n in enumerate(self.nodes):
                try:
                    _values[n] = value[i]
                except:
                    _values[n] = None
        elif isinstance(value, dict):
            for n in self.nodes:
                try:
                    _values[n] = value[n]
                except:
                    _values[n] = None
        else:
            log.error('Something went wrong, by formatting the node values!')
            raise CnetError
        return _values

    def format_edge_value(self, value):
        """Returns a dict with edge ids and assigned values."""
        # check if value is string, list or dict
        _values = {}
        if isinstance(value, str) or isinstance(value, int) or \
           isinstance(value, float) or isinstance(value, bool) \
           or isinstance(value, tuple):
            for n in self.edges:
                _values[n] = value
        elif isinstance(value, list):
            for i, n in enumerate(self.edges):
                try:
                    _values[n] = value[i]
                except:
                    _values[n] = None
        elif isinstance(value, dict):
            for n in self.edges:
                try:
                    _values[n] = value[n]
                except:
                    _values[n] = None
        else:
            log.error('Something went wrong, by formatting the edge values!')
            raise CnetError
        return _values

    def curve(self):
        """Calculate the bend factor for curved edges."""
        if 'edge_curved' in self.edge_attributes:
            _curved = {}
            for key, value in self.edge_attributes['edge_curved'].items():
                curved = value

                if curved == 0:
                    _curved[key] = 0
                else:
                    v1 = np.array([0, 0])
                    v2 = np.array([1, 1])
                    v3 = np.array([(2*v1[0]+v2[0]) / 3.0 - curved * 0.5 * (v2[1]-v1[1]),
                                   (2*v1[1]+v2[1]) / 3.0 +
                                   curved * 0.5 * (v2[0]-v1[0])
                                   ])
                    vec1 = v2-v1
                    vec2 = v3 - v1
                    angle = np.rad2deg(np.arccos(
                        np.dot(vec1, vec2) / np.sqrt((vec1*vec1).sum()) / np.sqrt((vec2*vec2).sum())))
                    _curved[key] = np.round(
                        np.sign(curved) * angle * -1, self.digits)
        return _curved


class TikzEdgeDrawer(object):
    """Class which handles the drawing of the edges.

    Parameters
    ----------
    id : edge id
        This parameter is the identifier (id) for the edge. Every edge should
        have a unique id.

    u : node id
        This parameter defines the origin of the edge (if directed), i.e. u->v.

    v : node id
        This parameter defines the destination of the edge (if directed)
        i.e. u->v.

    attr : keyword arguments, optional (default = no attributes)
        Attributes to add to edge as key=value pairs.
        See also :py:meth:`plot`

    See Also
    --------
    plot

    """

    def __init__(self, id, u, v, **attr):
        """Initialize the edge drawer.

        Parameters
        ----------
        id : edge id
            This parameter is the identifier (id) for the edge. Every edge
            should have a unique id.

        u : node id
            This parameter defines the origin of the edge (if directed).

        v : node id
            This parameter defines the destination of the edge (if directed).

        attr : keyword arguments, optional (default = no attributes)
            Attributes to add to edge as key=value pairs.
            See also :py:meth:`plot`
        """
        self.id = id
        self.u = u
        self.v = v
        self.attributes = attr
        self.digits = DIGITS
        # all options from the tikz-network library
        self.tikz_kwds = OrderedDict()
        self.tikz_kwds["edge_width"] = 'lw'
        self.tikz_kwds["edge_color"] = 'color'
        self.tikz_kwds["edge_r"] = 'R'
        self.tikz_kwds["edge_g"] = 'G'
        self.tikz_kwds["edge_b"] = 'B'
        self.tikz_kwds["edge_opacity"] = 'opacity'
        self.tikz_kwds["edge_curved"] = 'bend'
        self.tikz_kwds["edge_label"] = 'label'
        self.tikz_kwds["edge_label_position"] = 'position'
        self.tikz_kwds["edge_label_distance"] = 'distance'
        self.tikz_kwds["edge_label_color"] = 'fontcolor'
        self.tikz_kwds["edge_label_size"] = 'fontscale'
        self.tikz_kwds["edge_style"] = 'style'
        # self.tikz_kwds["edge_arrow_size"] = 'length'
        # self.tikz_kwds["edge_arrow_width"] = 'width'
        # self.tikz_kwds["edge_path"] = 'path'
        self.tikz_kwds["edge_loop_size"] = 'loopsize'
        self.tikz_kwds["edge_loop_position"] = 'loopposition'
        self.tikz_kwds["edge_loop_shape"] = 'loopshape'

        self.tikz_args = OrderedDict()
        self.tikz_args['edge_directed'] = 'Direct'
        self.tikz_args['edge_math_mode'] = 'Math'
        self.tikz_args['edge_rgb'] = 'RGB'
        self.tikz_args['edge_not_in_bg'] = 'NotInBG'

    def _check_color(self, mode='tex'):
        """Check if RGB colors are used and return this option."""
        _color = self.attributes.get('edge_color', None)
        if isinstance(_color, tuple) and mode == 'tex':
            self.attributes['edge_color'] = '{{{},{},{}}}'.format(
                _color[0], _color[1], _color[2])
            self.attributes['edge_rgb'] = True
        elif isinstance(_color, tuple) and mode == 'csv' and \
                self.attributes.get('edge_rgb', False):
            self.attributes['edge_color'] = None
            self.attributes['edge_r'] = _color[0]
            self.attributes['edge_g'] = _color[1]
            self.attributes['edge_b'] = _color[2]
        elif not isinstance(_color, tuple) and mode == 'csv' and \
                self.attributes.get('edge_rgb', False):
            self.attributes['edge_r'] = self.attributes.get('edge_r', 0)
            self.attributes['edge_g'] = self.attributes.get('edge_g', 0)
            self.attributes['edge_b'] = self.attributes.get('edge_b', 0)
        elif isinstance(_color, tuple) and mode == 'csv' and \
                self.attributes.get('edge_rgb', False) == False:
            self.attributes['edge_color'] = None

    def _format_style(self):
        """Format the style attribute for the edge.

        In order to change the arrow shape, the style attribute of the edge has
        to be changed. This option is only available for 'pdf' and 'tex'
        files. 'csv' files do not have this option.

        """
        if 'edge_arrow_size' in self.attributes:
            arrow_size = 'length=' + \
                str(15*self.attributes['edge_arrow_size'])+'cm,'
        else:
            arrow_size = ''

        if 'edge_arrow_size' in self.attributes:
            arrow_width = 'width=' + \
                str(10*self.attributes['edge_arrow_width'])+'cm'
        else:
            arrow_width = ''

        if (arrow_size != '' or arrow_width != '') and\
           self.attributes.get('edge_directed', False) == True:
            self.attributes['edge_style'] = '{{-{{Latex[{}{}]}}, {} }}'.format(
                arrow_size, arrow_width, self.attributes.get('edge_style', ''))

    def draw(self, mode='tex'):
        """Function to draw an virtual edge.

        Parameters
        ----------
        mode : str, optional (default = 'tex')
            The mode defines which kind of result should be returned. Currently
            a string for a tex file or a string for a csv file can be returned.

        Returns
        -------
        string : str
            Returns a string defining the edge. If 'tex' mode is enabled, a
            tikz-network code is returned. If 'csv' mode is enabled, a row
            element for the edge list is returned.

        """
        if mode == 'tex':
            self._format_style()
            self._check_color()

            string = '\\Edge['

            for k in self.tikz_kwds:
                if k in self.attributes and \
                   self.attributes.get(k, None) is not None:
                    string += ',{}={}'.format(self.tikz_kwds[k],
                                              self.attributes[k])
            for k in self.tikz_args:
                if k in self.attributes:
                    if self.attributes[k] == True:
                        string += ',{}'.format(self.tikz_args[k])

            string += ']({})({})'.format(self.u, self.v)

        elif mode == 'csv':
            self._check_color(mode='csv')
            string = '{},{}'.format(self.u, self.v)

            for k in self.tikz_kwds:
                if k in self.attributes:
                    if self.attributes[k] is not None:
                        string += ',{}'.format(self.attributes[k])
                    else:
                        string += ', '
            for k in self.tikz_args:
                if k in self.attributes:
                    if self.attributes[k] == True:
                        string += ',true'
                    else:
                        string += ',false'

        return string + '\n'

    def head(self):
        """Function to draw the header of an virtual edge.

        Returns
        -------
        string : str
            Returns a string with the attributes defined for the edge. This
            string can be used as header for the 'csv' file.

        """
        self._check_color(mode='csv')
        string = 'u,v'
        for k in self.tikz_kwds:
            if k in self.attributes:
                string += ',{}'.format(self.tikz_kwds[k])
        for k in self.tikz_args:
            if k in self.attributes:
                string += ',{}'.format(self.tikz_args[k])

        return string + '\n'


class TikzNodeDrawer(object):
    """Class which handles the drawing of the nodes

    Parameters
    ----------
    id : node id
        This parameter is the identifier (id) for the node. Every node should
        have a unique id.

    attr : keyword arguments, optional (default = no attributes)
        Attributes to add to node as key=value pairs.
        See also :py:meth:`plot`

    See Also
    --------
    plot

    """

    def __init__(self, id, **attr):
        """Initialize the node drawer.

        Parameters
        ----------
        id : node id
            This parameter is the identifier (id) for the node. Every node
            should have a unique id.

        attr : keyword arguments, optional (default = no attributes)
            Attributes to add to node as key=value pairs.
            See also :py:meth:`plot`

        """
        self.id = id
        self.x = attr.get('layout', (0, 0))[0]
        self.y = attr.get('layout', (0, 0))[1]
        self.attributes = attr
        self.digits = DIGITS
        # all options from the tikz-network library
        self.tikz_kwds = OrderedDict()
        self.tikz_kwds['node_size'] = 'size'
        self.tikz_kwds['node_color'] = 'color'
        self.tikz_kwds["node_r"] = 'R'
        self.tikz_kwds["node_g"] = 'G'
        self.tikz_kwds["node_b"] = 'B'
        self.tikz_kwds['node_opacity'] = 'opacity'
        self.tikz_kwds['node_label'] = 'label'
        self.tikz_kwds['node_label_position'] = 'position'
        self.tikz_kwds['node_label_distance'] = 'distance'
        self.tikz_kwds['node_label_color'] = 'fontcolor'
        self.tikz_kwds['node_label_size'] = 'fontscale'
        self.tikz_kwds['node_shape'] = 'shape'
        self.tikz_kwds['node_style'] = 'style'
        self.tikz_kwds['node_layer'] = 'layer'

        self.tikz_args = OrderedDict()
        self.tikz_args['node_label_off'] = 'NoLabel'
        self.tikz_args['node_label_as_id'] = 'IdAsLabel'
        self.tikz_args['node_math_mode'] = 'Math'
        self.tikz_args['node_rgb'] = 'RGB'
        self.tikz_args['node_pseudo'] = 'Pseudo'

    def _check_color(self, mode='tex'):
        """Check if RGB colors are used and return this option."""
        _color = self.attributes.get('node_color', None)
        if isinstance(_color, tuple) and mode == 'tex':
            self.attributes['node_color'] = '{{{},{},{}}}'.format(
                _color[0], _color[1], _color[2])
            self.attributes['node_rgb'] = True
        elif isinstance(_color, tuple) and mode == 'csv' and \
                self.attributes.get('node_rgb', False):
            self.attributes['node_color'] = None
            self.attributes['node_r'] = _color[0]
            self.attributes['node_g'] = _color[1]
            self.attributes['node_b'] = _color[2]
        elif not isinstance(_color, tuple) and mode == 'csv' and \
                self.attributes.get('node_rgb', False):
            self.attributes['node_r'] = self.attributes.get('node_r', 0)
            self.attributes['node_g'] = self.attributes.get('node_g', 0)
            self.attributes['node_b'] = self.attributes.get('node_b', 0)
        elif isinstance(_color, tuple) and mode == 'csv' and \
                self.attributes.get('node_rgb', False) == False:
            self.attributes['node_color'] = None

    def draw(self, mode='tex'):
        """Function to draw a virtual node.

        Parameters
        ----------
        mode : str, optional (default = 'tex')
            The mode defines which kind of result should be returned. Currently
            a string for a tex file or a string for a csv file can be returned.

        Returns
        -------
        string : str
            Returns a string defining the node. If 'tex' mode is enabled, a
            tikz-network code is returned. If 'csv' mode is enabled, a row
            element for the node list is returned.

        """
        if mode == 'tex':
            self._check_color()
            string = '\\Vertex[x={x:.{n}f},y={y:.{n}f}'\
                     ''.format(x=self.x, y=self.y, n=self.digits)

            for k in self.tikz_kwds:
                if k in self.attributes and \
                   self.attributes.get(k, None) is not None:
                    string += ',{}={}'.format(self.tikz_kwds[k],
                                              self.attributes[k])
            for k in self.tikz_args:
                if k in self.attributes:
                    if self.attributes[k] == True:
                        string += ',{}'.format(self.tikz_args[k])

            string += ']{{{}}}'.format(self.id)

        elif mode == 'csv':
            self._check_color(mode='csv')
            string = '{id},{x:.{n}f},{y:.{n}f}'\
                     ''.format(id=self.id, x=self.x, y=self.y, n=self.digits)

            for k in self.tikz_kwds:
                if k in self.attributes:
                    if self.attributes[k] is not None:
                        string += ',{}'.format(self.attributes[k])
                    else:
                        string += ', '

            for k in self.tikz_args:
                if k in self.attributes:
                    if self.attributes[k] == True:
                        string += ',true'
                    else:
                        string += ',false'

        return string + '\n'

    def head(self):
        """Function to draw the header of a virtual node.

        Returns
        -------
        string : str
            Returns a string with the attributes defined for the node. This
            string can be used as header for the 'csv' file.

        """
        self._check_color(mode='csv')
        string = 'id,x,y'
        for k in self.tikz_kwds:
            if k in self.attributes:
                string += ',{}'.format(self.tikz_kwds[k])
        for k in self.tikz_args:
            if k in self.attributes:
                string += ',{}'.format(self.tikz_args[k])

        return string + '\n'

# =============================================================================
# eof
#
# Local Variables:
# mode: python
# mode: linum
# mode: auto-fill
# fill-column: 80
# End:



#!/usr/bin/python -tt
# -*- coding: utf-8 -*-
# =============================================================================
# File      : canvas.py
# Creation  : 19 May 2018
# Time-stamp: <Don 2018-07-26 16:36 juergen>
#
# Copyright (c) 2018 Jürgen Hackl <hackl@ibi.baug.ethz.ch>
#               http://www.ibi.ethz.ch
# $Id$
#
# Description : Module to create a canvas for plotting
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# =============================================================================
from . import logger
from .exceptions import CnetError
log = logger(__name__)

# TODO: move to config file
CANVAS = (6, 6)


class Canvas(object):
    """A canvas object defining the size of the plot.

    Parameters
    ----------
    width : int or float, optional (default = 6)
        The parameter defines the width of the figure. The width is defined in
        cm units. If no width is defined the default value of 6 cm is used. The
        default can be changed in the config file.

    height : int or float, optional (default = 6)
        The parameter defines the height of the figure. The height is defined in
        cm units. If no height is defined the default value of 6 cm is used. The
        default can be changed in the config file.

    margins : None, int, float or dict, optional (default = None)
        The margins define the 'empty' space from the canvas border. If no
        margins are defined, the margin will be calculated based on the maximum
        node size, to avoid clipping of the nodes. If a single int or float is
        defined all margins using this distances. To define different the margin
        sizes for all size a dictionary with in the form of
        `{'top':2,'left':1,'bottom':2,'right':.5}` has to be used.

    node_sizes : dict, optional (default = None)
        If no specific margins are defined, the margins are based on the maximum
        node size, in order to avoid clipping at the boundary of the figure. If
        no node_sizes are defined, the default value of tikz-network will be
        used.

    Attributes
    ----------
    width : int or float
        Width of the figure. This property can be called, set and modified.

    height : int or float
        Height of the figure. This property can be called, set and modified.

    """

    def __init__(self, width=None, height=None, margins=None, node_sizes=None):
        """Initialize a canvas object.

        Parameters
        ----------
        width : int or float, optional (default = 6)
            The parameter defines the width of the figure. The width is defined
            in cm units. If no width is defined the default value of 6 cm is
            used. The default can be changed in the config file.

        height : int or float, optional (default = 6)
            The parameter defines the height of the figure. The height is
            defined in cm units. If no height is defined the default value of 6
            cm is used. The default can be changed in the config file.

        margins : None, int, float or dict, optional (default = None)
            The margins define the 'empty' space from the canvas border. If no
            margins are defined, the margin will be calculated based on the
            maximum node size, to avoid clipping of the nodes. If a single int
            or float is defined all margins using this distances. To define
            different the margin sizes for all size a dictionary with in the
            form of `{'top':2,'left':1,'bottom':2,'right':.5}` has to be used.

        node_sizes : dict, optional (default = None)
            If no specific margins are defined, the margins are based on the
            maximum node size, in order to avoid clipping at the boundary of the
            figure. If no node_sizes are defined, the default value of
            tikz-network will be used.

        """
        # apply default values if not defined
        if width is None:
            width = CANVAS[0]
        if height is None:
            height = CANVAS[1]

        # initialize variables
        self._width = width
        self._height = height
        self._margins = margins
        self._node_sizes = node_sizes

    @property
    def width(self):
        """Returns the width of the canvas."""
        return self._width

    @width.setter
    def width(self, width):
        """Set the width of the canvas."""
        self._width = width

    @property
    def height(self):
        """Returns the height of the canvas."""
        return self._height

    @height.setter
    def height(self, height):
        """Set the height of the canvas."""
        self._height = height

    def margins(self, margins=None, node_sizes=None):
        """Returns a dictionary of margins.

        Parameters
        ----------
        margins : None, int, float or dict, optional (default = None)
            The margins define the 'empty' space from the canvas border. If no
            margins are defined, the margin will be calculated based on the
            maximum node size, to avoid clipping of the nodes. If a single int
            or float is defined all margins using this distances. To define
            different the margin sizes for all size a dictionary with in the
            form of `{'top':2,'left':1,'bottom':2,'right':.5}` has to be used.

        node_sizes : dict, optional (default = None)
            If no specific margins are defined, the margins are based on the
            maximum node size, in order to avoid clipping at the boundary of the
            figure. If no node_sizes are defined, the default value of
            tikz-network will be used.

        Returns
        -------
        margins : dict
            Returns the margins of the :py:class:`Canvas` as a dictionary in
            from of `{'top':2,'left':1,'bottom':2,'right':.5}`. The values
            correspond to cm units.

        Examples
        --------
        Default margins if nothing is defined.

        >>> canvas = cn.Canvas()
        >>> canvas.margins()
        {'top':.35,'left':.35,'bottom':.35,'right':.35}

        Margins defined with one value

        >>> canvas.margins(1)
        {'top':1,'left':1,'bottom':1,'right':1}

        Margins as dictionary.

        >>> canvas.margins({'top':2,'left':1,'bottom':2,'right':.5})
        {'top':2,'left':1,'bottom':2,'right':.5}

        Margins defined via max node sizes:

        >>> canvas.margins(margins=None,node_sizes{'a':.3,'b':.5})
        {'top':.3,'left':.3,'bottom':.3,'right':.3}

        """

        # check if arguments are None
        if margins is None and self._margins is not None:
            margins = self._margins
        if node_sizes is None and self._node_sizes is not None:
            node_sizes = self._node_sizes

        # if margins are not defined use max node size to avoid clipping.
        if margins is None:
            if node_sizes is not None:
                _margin = max(node_sizes.values())/2+.05
            else:
                _margin = 0.35
            _margins = {'top': _margin, 'bottom': _margin,
                        'left': _margin, 'right': _margin}

        # if only one number is specified, this will be applied to all margins.
        elif isinstance(margins, int) or isinstance(margins, float):
            _m = margins
            _margins = {'top': _m, 'left': _m, 'bottom': _m, 'right': _m}

        # if margins defined as dict, the dict values are used
        elif isinstance(margins, dict):
            _margins = {'top': margins.get('top', 0),
                        'left': margins.get('left', 0),
                        'bottom': margins.get('bottom', 0),
                        'right': margins.get('right', 0)}
        else:
            log.error('Margins are not proper defined!')
            raise CnetError

        # check size of the margins
        if _margins['top'] + _margins['bottom'] >= self.height or \
           _margins['left'] + _margins['right'] >= self.width:
            log.error('Margins horizontal {} or vertical {} are larger than the'
                      ' canvas size ({},{})!'
                      ''.format(_margins['left'] + _margins['right'],
                                _margins['top'] + _margins['bottom'],
                                self.width, self.height))
            raise CnetError

        # update class variables
        self._margins = _margins
        self._node_sizes = node_sizes

        # return a dict of margins
        return _margins

    def fit(self, layout, keep_aspect_ratio=True):
        """Fit the node positions to the canvas.

        Parameters
        ----------
        layout : dict
            A dictionary with the node positions on a 2-dimensional plane. The
            key value of the dict represents the node id while the value
            represents a tuple of coordinates (e.g. n = (x,y)). The initial
            layout can be placed anywhere on the 2-dimensional plane.

        keep_aspect_ratio : bool, optional (default = True)
            Defines whether to keep the aspect ratio of the current layout. If
            `False`, the layout will be rescaled to fit exactly into the
            available area in the canvas (i.e. removed margins). If `True`, the
            original aspect ratio of the layout will be kept and it will be
            centered within the canvas.

        Returns
        -------
        layout : dict
            Returns a dictionary with the new node positions. Key values
            represents the node ids and the values are the new coordinates. The
            new coordinates are shifted and transformed from its origins.

        Examples
        --------
        Create empty canvas with no margins and fit simple layout.

        >>> canvas = cn.Canvas(6,4,margins=0)
        >>> layout = {'a':(-1,-1),'b':(1,-1),'c':(1,1),'d':(-1,1)}
        >>> canvas.fit(layout)
        {'a':(1,0),'b':(5,0),'c':(5,4),'d':(1,4)}

        Without keeping the aspect ratio.

        >>> canvas = cn.Canvas(6,4,margins=0)
        >>> layout = {'a':(-1,-1),'b':(1,-1),'c':(1,1),'d':(-1,1)}
        >>> canvas.fit(layout, keep_aspect_ratio=False)
        {'a':(0,0),'b':(6,0),'c':(6,4),'d':(0,4)}

        """
        # get canvas size and margins
        width = self.width
        height = self.height
        margins = self.margins()

        # find min and max values of the points
        min_x = min(layout.items(), key=lambda item: item[1][0])[1][0]
        max_x = max(layout.items(), key=lambda item: item[1][0])[1][0]
        min_y = min(layout.items(), key=lambda item: item[1][1])[1][1]
        max_y = max(layout.items(), key=lambda item: item[1][1])[1][1]

        # calculate the scaling ratio
        ratio_x = float('inf')
        ratio_y = float('inf')

        if max_x-min_x > 0:
            ratio_x = (width-margins['left']-margins['right']) / (max_x-min_x)
        if max_y-min_y > 0:
            ratio_y = (height-margins['top']-margins['bottom']) / (max_y-min_y)

        if keep_aspect_ratio:
            scaling = (min(ratio_x, ratio_y), min(ratio_x, ratio_y))
        else:
            scaling = (ratio_x, ratio_y)

        if scaling[0] == float('inf'):
            scaling = (1, scaling[1])
        if scaling[1] == float('inf'):
            scaling = (scaling[0], 1)

        # apply scaling to the points
        _layout = {}
        for n, (x, y) in layout.items():
            _x = (x)*scaling[0]
            _y = (y)*scaling[1]
            _layout[n] = (_x, _y)

        # find min and max values of new the points
        min_x = min(_layout.items(), key=lambda item: item[1][0])[1][0]
        max_x = max(_layout.items(), key=lambda item: item[1][0])[1][0]
        min_y = min(_layout.items(), key=lambda item: item[1][1])[1][1]
        max_y = max(_layout.items(), key=lambda item: item[1][1])[1][1]

        # calculate the translation
        translation = (((width-margins['left']-margins['right'])/2
                        + margins['left']) - ((max_x-min_x)/2 + min_x),
                       ((height-margins['top']-margins['bottom'])/2
                        + margins['bottom']) - ((max_y-min_y)/2 + min_y))

        # apply translation to the points
        for n, (x, y) in _layout.items():
            _x = (x)+translation[0]
            _y = (y)+translation[1]
            _layout[n] = (_x, _y)

        return _layout

# =============================================================================
# eof
#
# Local Variables:
# mode: python
# mode: linum
# mode: auto-fill
# fill-column: 80
# End:



#!/usr/bin/python -tt
# -*- coding: utf-8 -*-
# =============================================================================
# File      : diffusion.py -- Diffusion processes
# Author    : Juergen Hackl <hackl@ibi.baug.ethz.ch>
# Creation  : 2018-06-27
# Time-stamp: <Don 2018-07-05 08:35 juergen>
#
# Copyright (c) 2018 Juergen Hackl <hackl@ibi.baug.ethz.ch>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# =============================================================================

import numpy as np
from scipy import sparse
import scipy.sparse.linalg as sla
from cnet import logger
from cnet.utils.exceptions import CnetError

log = logger(__name__)


class RandomWalkDiffusion(object):
    """A class to simulate a 'normal' diffusion process using random walkers.

    Parameters
    ----------
    network : Network
       A :py:class:`Network` object.

    walkers : int, optional (default = 5)
       Number of random walkers

    epsilon : float, optional (default = 0.01)
       Acceptable difference between the simulation and the analytically
       solution

    maxiter : integer, optional (default = 1000)
       The maximum number of iterations used to find a the stationary
       distribution.

    """

    def __init__(self, network, walkers=5, epsilon=0.01, maxiter=1000, weight=None):

        self.network = network
        self.walkers = walkers
        self.epsilon = epsilon
        self.maxiter = maxiter

        self.T = self.network.transition_matrix(weight)
        self.pi = None

    def stationary_distribution(self, A=None, normalized=True, lanczos_vecs=15, maxiter=1000):
        """Compute normalized leading eigenvector of a given matrix A.

        Parameters
        ----------
        A : None or scipy.sparse.coo_matrix, optional (default = None)
            Sparse matrix for which leading eigenvector will be computed. If no
            matrix is defined the transition matrix of the network will be used.

        normalized: bool, optional (default =  True)
            It True the result will be normalized, otherwise not.

        lanczos_vecs: int, optional (default = 15) The number of Lanczos vectors
            is used in the approximate calculation of eigenvectors and
            eigenvalues. This maps to the ncv parameter scipy's underlying
            function eigs.

        maxiter: int, option (default = True) Scaling factor for the number of
            iterations to be used in the approximate calculation of eigenvectors
            and eigenvalues. The number of iterations passed to scipy's
            underlying eigs function will be n*maxiter where n is the number of
            rows/columns of the Laplacian matrix.

        Returns
        -------
        pi : numpy.array
           A vector of the leading eigenvector

        """
        if A is None:
            A = self.T.transpose()
        if not sparse.issparse(A):  # pragma: no cover
            log.error("The matrix A must be a sparse matrix!")
            raise CnetError

        # NOTE: ncv sets additional auxiliary eigenvectors that are computed
        # NOTE: in order to be more confident to find the one with the largest
        # NOTE: magnitude, see https://github.com/scipy/scipy/issues/4987
        w, pi = sla.eigs(A, k=1, which="LM", ncv=lanczos_vecs, maxiter=maxiter)
        pi = pi.reshape(pi.size, )
        if normalized:
            pi /= sum(pi)
        return pi

    def step(self, time, node=None):
        """Computes the diffusion process at a certain time step.

        Parameters
        ----------
        time: integer
           Time step at which the diffusion process should be evaluated.

        node: string, optional (default = None)
           Node id of the node where the diffusion process should be
           initialized. If no node id is defined a random node will be chosen
           instead.

        Returns
        -------
        p_t : dict
           A dictionary with the node ids as key values and the diffusion output
           as values.

        """
        T = self.T
        n = self.network.number_of_nodes()
        x = np.zeros(n)

        if node is None:
            x[np.random.randint(n)] = 1
        else:
            x[self.network.nodes.index(node)] = 1

        p_t = (T**time).transpose().dot(x)

        return {self.network.nodes[i].id: v for i, v in enumerate(p_t)}

    def speed(self):
        """ Computes the speed of the diffusion process.

        Computes the average number of steps requires by a random walk
        process to fall below a total variation distance below epsilon (TVD
        computed between the momentary visitation probabilities \pi^t and the
        stationary distribution \pi = \pi^{\infty}. This time can be used to
        measure diffusion speed in a given (weighted and directed) network.

        """
        T = self.T
        pi = self.stationary_distribution()
        n = self.network.number_of_nodes()

        avg_speed = []
        for w in range(self.walkers):
            _avg_speed = 0
            x = np.zeros(n)
            seed = np.random.randint(n)
            x[seed] = 1
            while self.TVD(x, pi) > self.epsilon:
                _avg_speed += 1
                x = T.transpose().dot(x)  # (T.dot(x.transpose())).transpose()
                if _avg_speed > self.maxiter:
                    raise RuntimeError(
                        'Maximal number of iteration was reached')
            avg_speed.append(_avg_speed)
        return np.mean(avg_speed)

    @staticmethod
    def TVD(p1, p2):
        """Compute total variation distance between two stochastic column vectors"""
        assert p1.shape == p2.shape
        return 0.5 * np.sum(np.absolute(np.subtract(p1, p2)))


# =============================================================================
# eof
#
# Local Variables:
# mode: python
# mode: linum
# mode: auto-fill
# fill-column: 80
# End:



"""This module contains the containers for the various classes
which are used to store nodes, edges and similar

To make the various classes pickle-able the defaultdicts need to be publicly addressable
function names, this means that no lambda functions are allowed.

All Pathpy classes which required a default value as a container, should use these here.
"""
from collections import defaultdict
import numpy as np


def nested_zero_default():
    """returns a nested default dict (2 levels)
    with a numpy zero array of length 0 as default
    """
    return defaultdict(zero_array_default)


def _zero_array():
    """returns a zero numpy array of length 2"""
    return np.array([0.0, 0.0])


def zero_array_default():
    """returns a default dict with numpy zero array af length 2 as default"""
    return defaultdict(_zero_array)



# -*- coding: utf-8 -*-
#    pathpy is an OpenSource python package for the analysis of time series data
#    on networks using higher- and multi order graphical models.
#
#    Copyright (C) 2016-2017 Ingo Scholtes, ETH Zürich
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    Contact the developer:

#    E-mail: ischoltes@ethz.ch
#    Web:    http://www.ingoscholtes.net

from pathpy import Paths
from pathpy.utils import Log
from pathpy.algorithms.shortest_paths import shortest_paths

import numpy as np


__all__ = ["paths_from_origin_destination", "read_origin_destination"]


def paths_from_origin_destination(origin_destination_list, network,
                                  distribute_weight=True):
    """extracts shortest path statistics based on origin/destination data.
    Such data capture the statistics of the origin (i.e. the start node) and destination
    (i.e. the target) node of itineraries in a given network.

    Common examples include passenger origin and destination statistics in transportation
    networks. The methods in this class can be used to read  origin/destination data from
    a file and generate path statistics based on the assumption that all paths from an
    origin and a destination follow the shortest path in the network.

    Extracts pathway statistics by calculating shortest paths between all origin and
    destination pairs in a given network.

    Parameters
    ----------
    origin_destination_list: list
        a list of tuples (o, d, w) containing the origin (o), destination (d),
        and float weight (w) of paths.
    network:
        the network topology for which shortest paths will be calculated. Names of nodes
        in the network must match the node names used in the origin destination list.
    distribute_weight: bool


    Returns
    -------
    paths
    """
    assert network is not None, \
        'Error: extraction of origin destination paths requires a network topology'

    all_paths = shortest_paths(network)

    paths = Paths()

    # OD is a list of tuples of the form (origin_node, destination_node, weight)
    # that indicates that the shortest path from origin_node to destination_node was
    # observed weight times
    Log.add('Starting origin destination path calculation ...')
    for (o, d, w) in origin_destination_list:
        assert o in network.nodes, 'Error: could not find node ' + str(o) + ' in network'
        assert d in network.nodes, 'Error: could not find node ' + str(d) + ' in network'
        sp = list(all_paths[o][d])
        num_paths = len(sp)
        if distribute_weight and num_paths > 1:
            # to avoid introducing false correlations that are not justified by the
            # available data, the (integer) weight of an origin destination pair can be
            # distributed among all possible shortest paths between a pair of nodes,
            # while constraining the weight of shortest paths to integers.
            for i in range(int(w)):
                paths.add_path_tuple(sp[i % num_paths], frequency=(0, 1))
        else:
            # in this case, the full weight of an origin destination path will be
            # assigned to a random single shortest path in the network
            paths.add_path_tuple(sp[np.random.randint(num_paths)], frequency=(0, w))
    Log.add('finished.')
    return paths


def read_origin_destination(filename, separator=','):
    """Helper function to read origin/destination statistics from a csv file.
    The file is assumed to have the following structure:

    origin1,destination1,weight
    origin2,destination2,weight
    origin3,destination3,weight

    Parameters
    ----------
    filename: str
        the path to the file from which to reach origin/destination statistics
    separator: str
        arbitrary separation character (default: ',')

    Returns
    -------
    list
    """
    origin_destination_list = []
    Log.add('Reading origin/destination statistics from file ...')

    with open(filename, 'r') as f:
        line = f.readline()
        while line:
            fields = line.rstrip().split(separator)
            origin_destination_list.append((fields[0].strip(), fields[1].strip(), float(fields[2].strip())))
            line = f.readline()
    Log.add('Finished.')

    return origin_destination_list




from pathpy.classes.higher_order_network import HigherOrderNetwork
from pathpy.classes.paths import Paths
from pathpy.classes.network import Network

import collections as _co

import numpy as _np



def _flow_net(paths, focal_node, self_loops=True):

    n = Network(directed=True)

    # consider all (sub-)paths of length two 
    # through the focal node
    for p in paths.paths[2]:
        if p[1] == focal_node:
            if self_loops or (p[0] != focal_node and p[2] != focal_node):
                src = 'src_{0}'.format(p[0])
                tgt = 'tgt_{0}'.format(p[2])
                mem = 'mem_{0}_{1}'.format(p[0], p[1])
                # calculate frequency of sub-paths src->focal_node->*, i.e. paths that 
                # continue through the focal_node
                # w_1 = 0
                # for x in paths.nodes:
                #         ct = p[:2]+(x,)
                #         if ct in paths.paths[2] and x != focal_node:
                #             w_1 += paths.paths[2][ct].sum()

                # calculate frequency of (sub-)path src -> focal_node -> tgt
                w_2 = paths.paths[2][p].sum()
                n.add_edge(src, mem, weight = 1)
                n.add_edge(mem, tgt, weight = w_2)


    # adjust weights of links to memory nodes:
    for m in n.nodes:
        if m.startswith('mem'):
            for u in n.predecessors[m]:
                n.edges[(u,m)]['weight'] = n.nodes[m]['outweight']
                n.nodes[m]['inweight'] = n.nodes[m]['outweight']
    return n


def _flow_net_markov(network, focal_node, self_loops=True):
    n = Network(directed=True)

    out_weight = _np.sum(network.nodes[focal_node]['outweight'])

    for u in network.predecessors[focal_node]:
        for w in network.successors[focal_node]:
            if self_loops or (u!= focal_node and w != focal_node):
                src = 'src_{0}'.format(u)
                tgt = 'tgt_{0}'.format(w)
                mem = 'mem_{0}_{1}'.format(u, focal_node)

                w_1 = _np.sum(network.edges[(u, focal_node)]['weight'])

                # at random, we expect the flow to be proportional to the relative edge weight
                w_2 = w_1 * (_np.sum(network.edges[(focal_node, w)]['weight'])/out_weight)
                n.add_edge(src, mem, weight=w_1)
                n.add_edge(mem, tgt, weight=w_2)
    return n


def _to_html(paths, focal_node, self_loops=True, markov=False, width=600, height=600, template_file=None):
    import json
    import os
    from string import Template

    import string
    import random

    if markov:
        g1 = HigherOrderNetwork(paths, k=1)
        n = _flow_net_markov(g1, focal_node=focal_node, self_loops=self_loops)
    else:
        n = _flow_net(paths, focal_node, self_loops=self_loops)

    
    node_idx = {}
    i = 0
    for v in n.nodes:
        node_idx[v] = i
        i += 1

    data = {
        'nodes': [ {'name': v, 'id': v} for v in n.nodes ],
        'links': [ {'source': int(node_idx[e[0]]), 'target': int(node_idx[e[1]]), 'value': n.edges[e]['weight']} for e in n.edges ]
    }

    div_id = "".join(random.choice(string.ascii_letters) for x in range(8))

    if template_file is None:
        module_dir = os.path.dirname(os.path.realpath(__file__))
        html_dir = os.path.join(module_dir, os.path.pardir, 'html_templates')            
        template_file = os.path.join(html_dir, 'alluvial_node.html') 

    with open(template_file) as f:
        html_str = f.read()

    args = {
        'flow_data': json.dumps(data),        
        'width': width,
        'height': height,
        'div_id': div_id,
        'focal_node': focal_node
    }

    # replace all placeholders in template
    return Template(html_str).substitute(args)


def write_html_flow(paths, focal_node, filename, markov=False, self_loops=True, width=600, height=600, template_file=None):
    html = _to_html(paths, focal_node, self_loops, markov, width, height, template_file)
    with open(filename, 'w+') as f:
        f.write(html)


def show_flow(paths, focal_node, markov=False, self_loops=True, width=600, height=600):
    html = _to_html(paths, focal_node, self_loops, markov, width, height)
    from IPython.core.display import display, HTML
    display(HTML(html))


def diffusion_to_flow_net(paths, markov=True, initial_node=None, steps=5):

    g1 = HigherOrderNetwork(paths, k=1)
    map_1 = g1.node_to_name_map()

    prob = _np.zeros(g1.vcount())
    prob = prob.transpose()
    if initial_node is None:
        initial_node = g1.nodes[0]    
    
    prob[map_1[initial_node]] = 1.0
    
    T = g1.transition_matrix()

    flow_net = Network(directed=True)

    if markov:
        # if markov == True flows are given by first-order transition matrix
        for t in range(1, steps+1):
            # calculate flow from i to j in step t
            for i in g1.nodes:
                for j in g1.nodes:                    
                    i_to_j = prob[map_1[i]] * T[map_1[j], map_1[i]]
                    if i_to_j > 0:
                        flow_net.add_edge('{0}_{1}'.format(i, t-1), '{0}_{1}'.format(j, t), weight = i_to_j)
            prob = T.dot(prob)
    else:
        # if markov == False calculate flows based on paths starting in initial_node
        for p in paths.paths[steps]:
            if p[0] == initial_node:
                for t in range(len(p)-1):
                    flow_net.add_edge('{0}_{1}'.format(p[t], t), '{0}_{1}'.format(p[t+1], t+1), weight = paths.paths[steps][p].sum())
        
        # normalize flows and balance in- and out-weight for all nodes
        # normalization = flow_net.nodes['{0}_{1}'.format(initial_node, 0)]['outweight']

        flow_net.nodes[initial_node+'_0']['inweight'] = 1.0
        Q = [initial_node+'_0']
        # adjust weights using BFS
        while Q:
            v = Q.pop()
            # print(v)
            inweight = flow_net.nodes[v]['inweight']
            outweight = flow_net.nodes[v]['outweight']

            for w in flow_net.successors[v]:
                flow_net.nodes[w]['inweight'] = flow_net.nodes[w]['inweight'] - flow_net.edges[(v,w)]['weight']
                flow_net.edges[(v,w)]['weight'] = (inweight/outweight) * flow_net.edges[(v,w)]['weight']
                flow_net.nodes[w]['inweight'] =  flow_net.nodes[w]['inweight'] + flow_net.edges[(v,w)]['weight']
                Q.append(w)
    return flow_net


def diffusion_to_html(paths, markov=True, initial_node=None, steps=5, width=600, height=600, template_file=None):
    import json
    import os
    from string import Template

    import string
    import random

    n = diffusion_to_flow_net(paths, markov=markov, initial_node=initial_node, steps=steps)

    node_map = {v: idx for idx, v in enumerate(n.nodes)}

    data = {
        'nodes': [ {'name': v, 'id': v} for v in n.nodes ],
        'links': [ {'source': node_map[e[0]], 'target': node_map[e[1]], 'value': n.edges[e]['weight']} for e in n.edges ]
    }

    div_id = "".join(random.choice(string.ascii_letters) for x in range(8))

    if template_file is None:
        module_dir = os.path.dirname(os.path.realpath(__file__))
        html_dir = os.path.join(module_dir, os.path.pardir, 'html_templates')            
        template_file = os.path.join(html_dir, 'alluvial_diffusion.html') 

    with open(template_file) as f:
        html_str = f.read()

    args = {
        'flow_data': json.dumps(data),
        'width': width,
        'height': height,
        'div_id': div_id
    }

    # replace all placeholders in template
    return Template(html_str).substitute(args)


def show_diffusion(paths, markov=True, initial_node=None, steps=5, width=600, height=600):
    html = diffusion_to_html(paths, markov=markov, initial_node=initial_node, steps=steps, width=width, height=height)
    from IPython.core.display import display, HTML
    display(HTML(html))


def write_html_diffusion(paths, filename, markov=True, initial_node=None, steps=5, width=600, height=600, template_file=None):
    html = diffusion_to_html(paths, markov=markov, initial_node=initial_node, steps=steps, width=width, height=height, template_file=template_file)
    with open(filename, 'w+') as f:
        f.write(html)